#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""

import datetime
from enum import Enum
from .timeutils import Calendar


class Holiday(Enum):
    """中国内地法定节假日枚举类"""

    def __new__(cls, english, chinese, days):
        """重写__new__方法，使枚举类的实例可以存储中文名和假期天数
        Args:
            english: 节假日英文名
            chinese: 节假日中文名
            days: 节假日天数
        """
        obj = object.__new__(cls)
        obj._value_ = english

        obj.chinese = chinese
        obj.days = days
        return obj

    new_years_day = "New Year's Day", "元旦", 1
    spring_festival = "Spring Festival", "春节", 3
    tomb_sweeping_day = "Tomb-sweeping Day", "清明", 1
    labour_day = "Labour Day", "劳动节", 1
    dragon_boat_festival = "Dragon Boat Festival", "端午", 1
    national_day = "National Day", "国庆节", 3
    mid_autumn_festival = "Mid-autumn Festival", "中秋", 1

    # special holidays
    anti_fascist_70th_day = "Anti-Fascist 70th Day", "中国人民抗日战争暨世界反法西斯战争胜利70周年纪念日", 1


class ChineseCalendar(Calendar):
    """中国内地交易日历类"""
    holidays_dict = {
        datetime.date(year=2004, month=1, day=1): Holiday.new_years_day.value,
        datetime.date(year=2004, month=1, day=22): Holiday.spring_festival.value,
        datetime.date(year=2004, month=1, day=23): Holiday.spring_festival.value,
        datetime.date(year=2004, month=1, day=24): Holiday.spring_festival.value,
        datetime.date(year=2004, month=1, day=25): Holiday.spring_festival.value,
        datetime.date(year=2004, month=1, day=26): Holiday.spring_festival.value,
        datetime.date(year=2004, month=1, day=27): Holiday.spring_festival.value,
        datetime.date(year=2004, month=1, day=28): Holiday.spring_festival.value,
        datetime.date(year=2004, month=5, day=1): Holiday.labour_day.value,
        datetime.date(year=2004, month=5, day=2): Holiday.labour_day.value,
        datetime.date(year=2004, month=5, day=3): Holiday.labour_day.value,
        datetime.date(year=2004, month=5, day=4): Holiday.labour_day.value,
        datetime.date(year=2004, month=5, day=5): Holiday.labour_day.value,
        datetime.date(year=2004, month=5, day=6): Holiday.labour_day.value,
        datetime.date(year=2004, month=5, day=7): Holiday.labour_day.value,
        datetime.date(year=2004, month=10, day=1): Holiday.national_day.value,
        datetime.date(year=2004, month=10, day=2): Holiday.national_day.value,
        datetime.date(year=2004, month=10, day=3): Holiday.national_day.value,
        datetime.date(year=2004, month=10, day=4): Holiday.national_day.value,
        datetime.date(year=2004, month=10, day=5): Holiday.national_day.value,
        datetime.date(year=2004, month=10, day=6): Holiday.national_day.value,
        datetime.date(year=2004, month=10, day=7): Holiday.national_day.value,
        datetime.date(year=2005, month=1, day=1): Holiday.new_years_day.value,
        datetime.date(year=2005, month=1, day=2): Holiday.new_years_day.value,
        datetime.date(year=2005, month=1, day=3): Holiday.new_years_day.value,
        datetime.date(year=2005, month=2, day=9): Holiday.spring_festival.value,
        datetime.date(year=2005, month=2, day=10): Holiday.spring_festival.value,
        datetime.date(year=2005, month=2, day=11): Holiday.spring_festival.value,
        datetime.date(year=2005, month=2, day=12): Holiday.spring_festival.value,
        datetime.date(year=2005, month=2, day=13): Holiday.spring_festival.value,
        datetime.date(year=2005, month=2, day=14): Holiday.spring_festival.value,
        datetime.date(year=2005, month=2, day=15): Holiday.spring_festival.value,
        datetime.date(year=2005, month=5, day=1): Holiday.labour_day.value,
        datetime.date(year=2005, month=5, day=2): Holiday.labour_day.value,
        datetime.date(year=2005, month=5, day=3): Holiday.labour_day.value,
        datetime.date(year=2005, month=5, day=4): Holiday.labour_day.value,
        datetime.date(year=2005, month=5, day=5): Holiday.labour_day.value,
        datetime.date(year=2005, month=5, day=6): Holiday.labour_day.value,
        datetime.date(year=2005, month=5, day=7): Holiday.labour_day.value,
        datetime.date(year=2005, month=10, day=1): Holiday.national_day.value,
        datetime.date(year=2005, month=10, day=2): Holiday.national_day.value,
        datetime.date(year=2005, month=10, day=3): Holiday.national_day.value,
        datetime.date(year=2005, month=10, day=4): Holiday.national_day.value,
        datetime.date(year=2005, month=10, day=5): Holiday.national_day.value,
        datetime.date(year=2005, month=10, day=6): Holiday.national_day.value,
        datetime.date(year=2005, month=10, day=7): Holiday.national_day.value,
        datetime.date(year=2006, month=1, day=1): Holiday.new_years_day.value,
        datetime.date(year=2006, month=1, day=2): Holiday.new_years_day.value,
        datetime.date(year=2006, month=1, day=3): Holiday.new_years_day.value,
        datetime.date(year=2006, month=1, day=29): Holiday.spring_festival.value,
        datetime.date(year=2006, month=1, day=30): Holiday.spring_festival.value,
        datetime.date(year=2006, month=1, day=31): Holiday.spring_festival.value,
        datetime.date(year=2006, month=2, day=1): Holiday.spring_festival.value,
        datetime.date(year=2006, month=2, day=2): Holiday.spring_festival.value,
        datetime.date(year=2006, month=2, day=3): Holiday.spring_festival.value,
        datetime.date(year=2006, month=2, day=4): Holiday.spring_festival.value,
        datetime.date(year=2006, month=5, day=1): Holiday.labour_day.value,
        datetime.date(year=2006, month=5, day=2): Holiday.labour_day.value,
        datetime.date(year=2006, month=5, day=3): Holiday.labour_day.value,
        datetime.date(year=2006, month=5, day=4): Holiday.labour_day.value,
        datetime.date(year=2006, month=5, day=5): Holiday.labour_day.value,
        datetime.date(year=2006, month=5, day=6): Holiday.labour_day.value,
        datetime.date(year=2006, month=5, day=7): Holiday.labour_day.value,
        datetime.date(year=2006, month=10, day=1): Holiday.national_day.value,
        datetime.date(year=2006, month=10, day=2): Holiday.national_day.value,
        datetime.date(year=2006, month=10, day=3): Holiday.national_day.value,
        datetime.date(year=2006, month=10, day=4): Holiday.national_day.value,
        datetime.date(year=2006, month=10, day=5): Holiday.national_day.value,
        datetime.date(year=2006, month=10, day=6): Holiday.national_day.value,
        datetime.date(year=2006, month=10, day=7): Holiday.national_day.value,
        datetime.date(year=2007, month=1, day=1): Holiday.new_years_day.value,
        datetime.date(year=2007, month=1, day=2): Holiday.new_years_day.value,
        datetime.date(year=2007, month=1, day=3): Holiday.new_years_day.value,
        datetime.date(year=2007, month=2, day=18): Holiday.spring_festival.value,
        datetime.date(year=2007, month=2, day=19): Holiday.spring_festival.value,
        datetime.date(year=2007, month=2, day=20): Holiday.spring_festival.value,
        datetime.date(year=2007, month=2, day=21): Holiday.spring_festival.value,
        datetime.date(year=2007, month=2, day=22): Holiday.spring_festival.value,
        datetime.date(year=2007, month=2, day=23): Holiday.spring_festival.value,
        datetime.date(year=2007, month=2, day=24): Holiday.spring_festival.value,
        datetime.date(year=2007, month=5, day=1): Holiday.labour_day.value,
        datetime.date(year=2007, month=5, day=2): Holiday.labour_day.value,
        datetime.date(year=2007, month=5, day=3): Holiday.labour_day.value,
        datetime.date(year=2007, month=5, day=4): Holiday.labour_day.value,
        datetime.date(year=2007, month=5, day=5): Holiday.labour_day.value,
        datetime.date(year=2007, month=5, day=6): Holiday.labour_day.value,
        datetime.date(year=2007, month=5, day=7): Holiday.labour_day.value,
        datetime.date(year=2007, month=10, day=1): Holiday.national_day.value,
        datetime.date(year=2007, month=10, day=2): Holiday.national_day.value,
        datetime.date(year=2007, month=10, day=3): Holiday.national_day.value,
        datetime.date(year=2007, month=10, day=4): Holiday.national_day.value,
        datetime.date(year=2007, month=10, day=5): Holiday.national_day.value,
        datetime.date(year=2007, month=10, day=6): Holiday.national_day.value,
        datetime.date(year=2007, month=10, day=7): Holiday.national_day.value,
        datetime.date(year=2007, month=12, day=30): Holiday.new_years_day.value,
        datetime.date(year=2007, month=12, day=31): Holiday.new_years_day.value,
        datetime.date(year=2008, month=1, day=1): Holiday.new_years_day.value,
        datetime.date(year=2008, month=2, day=6): Holiday.spring_festival.value,
        datetime.date(year=2008, month=2, day=7): Holiday.spring_festival.value,
        datetime.date(year=2008, month=2, day=8): Holiday.spring_festival.value,
        datetime.date(year=2008, month=2, day=9): Holiday.spring_festival.value,
        datetime.date(year=2008, month=2, day=10): Holiday.spring_festival.value,
        datetime.date(year=2008, month=2, day=11): Holiday.spring_festival.value,
        datetime.date(year=2008, month=2, day=12): Holiday.spring_festival.value,
        datetime.date(year=2008, month=4, day=4): Holiday.tomb_sweeping_day.value,
        datetime.date(year=2008, month=4, day=5): Holiday.tomb_sweeping_day.value,
        datetime.date(year=2008, month=4, day=6): Holiday.tomb_sweeping_day.value,
        datetime.date(year=2008, month=5, day=1): Holiday.labour_day.value,
        datetime.date(year=2008, month=5, day=2): Holiday.labour_day.value,
        datetime.date(year=2008, month=5, day=3): Holiday.labour_day.value,
        datetime.date(year=2008, month=6, day=7): Holiday.dragon_boat_festival.value,
        datetime.date(year=2008, month=6, day=8): Holiday.dragon_boat_festival.value,
        datetime.date(year=2008, month=6, day=9): Holiday.dragon_boat_festival.value,
        datetime.date(year=2008, month=9, day=13): Holiday.mid_autumn_festival.value,
        datetime.date(year=2008, month=9, day=14): Holiday.mid_autumn_festival.value,
        datetime.date(year=2008, month=9, day=15): Holiday.mid_autumn_festival.value,
        datetime.date(year=2008, month=9, day=29): Holiday.national_day.value,
        datetime.date(year=2008, month=9, day=30): Holiday.national_day.value,
        datetime.date(year=2008, month=10, day=1): Holiday.national_day.value,
        datetime.date(year=2008, month=10, day=2): Holiday.national_day.value,
        datetime.date(year=2008, month=10, day=3): Holiday.national_day.value,
        datetime.date(year=2008, month=10, day=4): Holiday.national_day.value,
        datetime.date(year=2008, month=10, day=5): Holiday.national_day.value,
        datetime.date(year=2009, month=1, day=1): Holiday.new_years_day.value,
        datetime.date(year=2009, month=1, day=2): Holiday.new_years_day.value,
        datetime.date(year=2009, month=1, day=3): Holiday.new_years_day.value,
        datetime.date(year=2009, month=1, day=25): Holiday.spring_festival.value,
        datetime.date(year=2009, month=1, day=26): Holiday.spring_festival.value,
        datetime.date(year=2009, month=1, day=27): Holiday.spring_festival.value,
        datetime.date(year=2009, month=1, day=28): Holiday.spring_festival.value,
        datetime.date(year=2009, month=1, day=29): Holiday.spring_festival.value,
        datetime.date(year=2009, month=1, day=30): Holiday.spring_festival.value,
        datetime.date(year=2009, month=1, day=31): Holiday.spring_festival.value,
        datetime.date(year=2009, month=4, day=4): Holiday.tomb_sweeping_day.value,
        datetime.date(year=2009, month=4, day=5): Holiday.tomb_sweeping_day.value,
        datetime.date(year=2009, month=4, day=6): Holiday.tomb_sweeping_day.value,
        datetime.date(year=2009, month=5, day=1): Holiday.labour_day.value,
        datetime.date(year=2009, month=5, day=2): Holiday.labour_day.value,
        datetime.date(year=2009, month=5, day=3): Holiday.labour_day.value,
        datetime.date(year=2009, month=5, day=28): Holiday.dragon_boat_festival.value,
        datetime.date(year=2009, month=5, day=29): Holiday.dragon_boat_festival.value,
        datetime.date(year=2009, month=5, day=30): Holiday.dragon_boat_festival.value,
        datetime.date(year=2009, month=10, day=1): Holiday.national_day.value,
        datetime.date(year=2009, month=10, day=2): Holiday.national_day.value,
        datetime.date(year=2009, month=10, day=3): Holiday.mid_autumn_festival.value,
        datetime.date(year=2009, month=10, day=4): Holiday.national_day.value,
        datetime.date(year=2009, month=10, day=5): Holiday.national_day.value,
        datetime.date(year=2009, month=10, day=6): Holiday.national_day.value,
        datetime.date(year=2009, month=10, day=7): Holiday.national_day.value,
        datetime.date(year=2009, month=10, day=8): Holiday.national_day.value,
        datetime.date(year=2010, month=1, day=1): Holiday.new_years_day.value,
        datetime.date(year=2010, month=1, day=2): Holiday.new_years_day.value,
        datetime.date(year=2010, month=1, day=3): Holiday.new_years_day.value,
        datetime.date(year=2010, month=2, day=13): Holiday.spring_festival.value,
        datetime.date(year=2010, month=2, day=14): Holiday.spring_festival.value,
        datetime.date(year=2010, month=2, day=15): Holiday.spring_festival.value,
        datetime.date(year=2010, month=2, day=16): Holiday.spring_festival.value,
        datetime.date(year=2010, month=2, day=17): Holiday.spring_festival.value,
        datetime.date(year=2010, month=2, day=18): Holiday.spring_festival.value,
        datetime.date(year=2010, month=2, day=19): Holiday.spring_festival.value,
        datetime.date(year=2010, month=4, day=3): Holiday.tomb_sweeping_day.value,
        datetime.date(year=2010, month=4, day=4): Holiday.tomb_sweeping_day.value,
        datetime.date(year=2010, month=4, day=5): Holiday.tomb_sweeping_day.value,
        datetime.date(year=2010, month=5, day=1): Holiday.labour_day.value,
        datetime.date(year=2010, month=5, day=2): Holiday.labour_day.value,
        datetime.date(year=2010, month=5, day=3): Holiday.labour_day.value,
        datetime.date(year=2010, month=6, day=14): Holiday.dragon_boat_festival.value,
        datetime.date(year=2010, month=6, day=15): Holiday.dragon_boat_festival.value,
        datetime.date(year=2010, month=6, day=16): Holiday.dragon_boat_festival.value,
        datetime.date(year=2010, month=9, day=22): Holiday.mid_autumn_festival.value,
        datetime.date(year=2010, month=9, day=23): Holiday.mid_autumn_festival.value,
        datetime.date(year=2010, month=9, day=24): Holiday.mid_autumn_festival.value,
        datetime.date(year=2010, month=10, day=1): Holiday.national_day.value,
        datetime.date(year=2010, month=10, day=2): Holiday.national_day.value,
        datetime.date(year=2010, month=10, day=3): Holiday.national_day.value,
        datetime.date(year=2010, month=10, day=4): Holiday.national_day.value,
        datetime.date(year=2010, month=10, day=5): Holiday.national_day.value,
        datetime.date(year=2010, month=10, day=6): Holiday.national_day.value,
        datetime.date(year=2010, month=10, day=7): Holiday.national_day.value,
        datetime.date(year=2011, month=1, day=1): Holiday.new_years_day.value,
        datetime.date(year=2011, month=1, day=2): Holiday.new_years_day.value,
        datetime.date(year=2011, month=1, day=3): Holiday.new_years_day.value,
        datetime.date(year=2011, month=2, day=2): Holiday.spring_festival.value,
        datetime.date(year=2011, month=2, day=3): Holiday.spring_festival.value,
        datetime.date(year=2011, month=2, day=4): Holiday.spring_festival.value,
        datetime.date(year=2011, month=2, day=5): Holiday.spring_festival.value,
        datetime.date(year=2011, month=2, day=6): Holiday.spring_festival.value,
        datetime.date(year=2011, month=2, day=7): Holiday.spring_festival.value,
        datetime.date(year=2011, month=2, day=8): Holiday.spring_festival.value,
        datetime.date(year=2011, month=4, day=3): Holiday.tomb_sweeping_day.value,
        datetime.date(year=2011, month=4, day=4): Holiday.tomb_sweeping_day.value,
        datetime.date(year=2011, month=4, day=5): Holiday.tomb_sweeping_day.value,
        datetime.date(year=2011, month=4, day=30): Holiday.labour_day.value,
        datetime.date(year=2011, month=5, day=1): Holiday.labour_day.value,
        datetime.date(year=2011, month=5, day=2): Holiday.labour_day.value,
        datetime.date(year=2011, month=6, day=4): Holiday.dragon_boat_festival.value,
        datetime.date(year=2011, month=6, day=6): Holiday.dragon_boat_festival.value,
        datetime.date(year=2011, month=9, day=10): Holiday.mid_autumn_festival.value,
        datetime.date(year=2011, month=9, day=11): Holiday.mid_autumn_festival.value,
        datetime.date(year=2011, month=9, day=12): Holiday.mid_autumn_festival.value,
        datetime.date(year=2011, month=10, day=1): Holiday.national_day.value,
        datetime.date(year=2011, month=10, day=2): Holiday.national_day.value,
        datetime.date(year=2011, month=10, day=3): Holiday.national_day.value,
        datetime.date(year=2011, month=10, day=4): Holiday.national_day.value,
        datetime.date(year=2011, month=10, day=5): Holiday.national_day.value,
        datetime.date(year=2011, month=10, day=6): Holiday.national_day.value,
        datetime.date(year=2011, month=10, day=7): Holiday.national_day.value,
        datetime.date(year=2012, month=1, day=1): Holiday.new_years_day.value,
        datetime.date(year=2012, month=1, day=2): Holiday.new_years_day.value,
        datetime.date(year=2012, month=1, day=3): Holiday.new_years_day.value,
        datetime.date(year=2012, month=1, day=22): Holiday.spring_festival.value,
        datetime.date(year=2012, month=1, day=23): Holiday.spring_festival.value,
        datetime.date(year=2012, month=1, day=24): Holiday.spring_festival.value,
        datetime.date(year=2012, month=1, day=25): Holiday.spring_festival.value,
        datetime.date(year=2012, month=1, day=26): Holiday.spring_festival.value,
        datetime.date(year=2012, month=1, day=27): Holiday.spring_festival.value,
        datetime.date(year=2012, month=1, day=28): Holiday.spring_festival.value,
        datetime.date(year=2012, month=4, day=2): Holiday.tomb_sweeping_day.value,
        datetime.date(year=2012, month=4, day=3): Holiday.tomb_sweeping_day.value,
        datetime.date(year=2012, month=4, day=4): Holiday.tomb_sweeping_day.value,
        datetime.date(year=2012, month=4, day=29): Holiday.labour_day.value,
        datetime.date(year=2012, month=4, day=30): Holiday.labour_day.value,
        datetime.date(year=2012, month=5, day=1): Holiday.labour_day.value,
        datetime.date(year=2012, month=6, day=22): Holiday.dragon_boat_festival.value,
        datetime.date(year=2012, month=6, day=24): Holiday.dragon_boat_festival.value,
        datetime.date(year=2012, month=9, day=30): Holiday.mid_autumn_festival.value,
        datetime.date(year=2012, month=10, day=1): Holiday.national_day.value,
        datetime.date(year=2012, month=10, day=2): Holiday.national_day.value,
        datetime.date(year=2012, month=10, day=3): Holiday.national_day.value,
        datetime.date(year=2012, month=10, day=4): Holiday.national_day.value,
        datetime.date(year=2012, month=10, day=5): Holiday.national_day.value,
        datetime.date(year=2012, month=10, day=6): Holiday.national_day.value,
        datetime.date(year=2012, month=10, day=7): Holiday.national_day.value,
        datetime.date(year=2013, month=1, day=1): Holiday.new_years_day.value,
        datetime.date(year=2013, month=1, day=2): Holiday.new_years_day.value,
        datetime.date(year=2013, month=1, day=3): Holiday.new_years_day.value,
        datetime.date(year=2013, month=2, day=9): Holiday.spring_festival.value,
        datetime.date(year=2013, month=2, day=10): Holiday.spring_festival.value,
        datetime.date(year=2013, month=2, day=11): Holiday.spring_festival.value,
        datetime.date(year=2013, month=2, day=12): Holiday.spring_festival.value,
        datetime.date(year=2013, month=2, day=13): Holiday.spring_festival.value,
        datetime.date(year=2013, month=2, day=14): Holiday.spring_festival.value,
        datetime.date(year=2013, month=2, day=15): Holiday.spring_festival.value,
        datetime.date(year=2013, month=4, day=4): Holiday.tomb_sweeping_day.value,
        datetime.date(year=2013, month=4, day=5): Holiday.tomb_sweeping_day.value,
        datetime.date(year=2013, month=4, day=6): Holiday.tomb_sweeping_day.value,
        datetime.date(year=2013, month=4, day=29): Holiday.labour_day.value,
        datetime.date(year=2013, month=4, day=30): Holiday.labour_day.value,
        datetime.date(year=2013, month=5, day=1): Holiday.labour_day.value,
        datetime.date(year=2013, month=6, day=10): Holiday.dragon_boat_festival.value,
        datetime.date(year=2013, month=6, day=11): Holiday.dragon_boat_festival.value,
        datetime.date(year=2013, month=6, day=12): Holiday.dragon_boat_festival.value,
        datetime.date(year=2013, month=9, day=19): Holiday.mid_autumn_festival.value,
        datetime.date(year=2013, month=9, day=20): Holiday.mid_autumn_festival.value,
        datetime.date(year=2013, month=9, day=21): Holiday.mid_autumn_festival.value,
        datetime.date(year=2013, month=10, day=1): Holiday.national_day.value,
        datetime.date(year=2013, month=10, day=2): Holiday.national_day.value,
        datetime.date(year=2013, month=10, day=3): Holiday.national_day.value,
        datetime.date(year=2013, month=10, day=4): Holiday.national_day.value,
        datetime.date(year=2013, month=10, day=5): Holiday.national_day.value,
        datetime.date(year=2013, month=10, day=6): Holiday.national_day.value,
        datetime.date(year=2013, month=10, day=7): Holiday.national_day.value,
        datetime.date(year=2014, month=1, day=1): Holiday.new_years_day.value,
        datetime.date(year=2014, month=1, day=31): Holiday.spring_festival.value,
        datetime.date(year=2014, month=2, day=1): Holiday.spring_festival.value,
        datetime.date(year=2014, month=2, day=2): Holiday.spring_festival.value,
        datetime.date(year=2014, month=2, day=3): Holiday.spring_festival.value,
        datetime.date(year=2014, month=2, day=4): Holiday.spring_festival.value,
        datetime.date(year=2014, month=2, day=5): Holiday.spring_festival.value,
        datetime.date(year=2014, month=2, day=6): Holiday.spring_festival.value,
        datetime.date(year=2014, month=4, day=5): Holiday.tomb_sweeping_day.value,
        datetime.date(year=2014, month=4, day=6): Holiday.tomb_sweeping_day.value,
        datetime.date(year=2014, month=4, day=7): Holiday.tomb_sweeping_day.value,
        datetime.date(year=2014, month=5, day=1): Holiday.labour_day.value,
        datetime.date(year=2014, month=5, day=2): Holiday.labour_day.value,
        datetime.date(year=2014, month=5, day=3): Holiday.labour_day.value,
        datetime.date(year=2014, month=6, day=2): Holiday.dragon_boat_festival.value,
        datetime.date(year=2014, month=9, day=8): Holiday.mid_autumn_festival.value,
        datetime.date(year=2014, month=10, day=1): Holiday.national_day.value,
        datetime.date(year=2014, month=10, day=2): Holiday.national_day.value,
        datetime.date(year=2014, month=10, day=3): Holiday.national_day.value,
        datetime.date(year=2014, month=10, day=4): Holiday.national_day.value,
        datetime.date(year=2014, month=10, day=5): Holiday.national_day.value,
        datetime.date(year=2014, month=10, day=6): Holiday.national_day.value,
        datetime.date(year=2014, month=10, day=7): Holiday.national_day.value,
        datetime.date(year=2015, month=1, day=1): Holiday.new_years_day.value,
        datetime.date(year=2015, month=1, day=2): Holiday.new_years_day.value,
        datetime.date(year=2015, month=1, day=3): Holiday.new_years_day.value,
        datetime.date(year=2015, month=2, day=18): Holiday.spring_festival.value,
        datetime.date(year=2015, month=2, day=19): Holiday.spring_festival.value,
        datetime.date(year=2015, month=2, day=20): Holiday.spring_festival.value,
        datetime.date(year=2015, month=2, day=21): Holiday.spring_festival.value,
        datetime.date(year=2015, month=2, day=22): Holiday.spring_festival.value,
        datetime.date(year=2015, month=2, day=23): Holiday.spring_festival.value,
        datetime.date(year=2015, month=2, day=24): Holiday.spring_festival.value,
        datetime.date(year=2015, month=4, day=5): Holiday.tomb_sweeping_day.value,
        datetime.date(year=2015, month=4, day=6): Holiday.tomb_sweeping_day.value,
        datetime.date(year=2015, month=5, day=1): Holiday.labour_day.value,
        datetime.date(year=2015, month=6, day=20): Holiday.dragon_boat_festival.value,
        datetime.date(year=2015, month=6, day=22): Holiday.dragon_boat_festival.value,
        datetime.date(year=2015, month=9, day=3): Holiday.anti_fascist_70th_day.value,
        datetime.date(year=2015, month=9, day=4): Holiday.anti_fascist_70th_day.value,
        datetime.date(year=2015, month=9, day=27): Holiday.mid_autumn_festival.value,
        datetime.date(year=2015, month=10, day=1): Holiday.national_day.value,
        datetime.date(year=2015, month=10, day=2): Holiday.national_day.value,
        datetime.date(year=2015, month=10, day=3): Holiday.national_day.value,
        datetime.date(year=2015, month=10, day=4): Holiday.national_day.value,
        datetime.date(year=2015, month=10, day=5): Holiday.national_day.value,
        datetime.date(year=2015, month=10, day=6): Holiday.national_day.value,
        datetime.date(year=2015, month=10, day=7): Holiday.national_day.value,
        datetime.date(year=2016, month=1, day=1): Holiday.new_years_day.value,
        datetime.date(year=2016, month=2, day=7): Holiday.spring_festival.value,
        datetime.date(year=2016, month=2, day=8): Holiday.spring_festival.value,
        datetime.date(year=2016, month=2, day=9): Holiday.spring_festival.value,
        datetime.date(year=2016, month=2, day=10): Holiday.spring_festival.value,
        datetime.date(year=2016, month=2, day=11): Holiday.spring_festival.value,
        datetime.date(year=2016, month=2, day=12): Holiday.spring_festival.value,
        datetime.date(year=2016, month=2, day=13): Holiday.spring_festival.value,
        datetime.date(year=2016, month=4, day=4): Holiday.tomb_sweeping_day.value,
        datetime.date(year=2016, month=5, day=1): Holiday.labour_day.value,
        datetime.date(year=2016, month=5, day=2): Holiday.labour_day.value,
        datetime.date(year=2016, month=6, day=9): Holiday.dragon_boat_festival.value,
        datetime.date(year=2016, month=6, day=10): Holiday.dragon_boat_festival.value,
        datetime.date(year=2016, month=6, day=11): Holiday.dragon_boat_festival.value,
        datetime.date(year=2016, month=9, day=15): Holiday.mid_autumn_festival.value,
        datetime.date(year=2016, month=9, day=16): Holiday.mid_autumn_festival.value,
        datetime.date(year=2016, month=9, day=17): Holiday.mid_autumn_festival.value,
        datetime.date(year=2016, month=10, day=1): Holiday.national_day.value,
        datetime.date(year=2016, month=10, day=2): Holiday.national_day.value,
        datetime.date(year=2016, month=10, day=3): Holiday.national_day.value,
        datetime.date(year=2016, month=10, day=4): Holiday.national_day.value,
        datetime.date(year=2016, month=10, day=5): Holiday.national_day.value,
        datetime.date(year=2016, month=10, day=6): Holiday.national_day.value,
        datetime.date(year=2016, month=10, day=7): Holiday.national_day.value,
        datetime.date(year=2017, month=1, day=1): Holiday.new_years_day.value,
        datetime.date(year=2017, month=1, day=2): Holiday.new_years_day.value,
        datetime.date(year=2017, month=1, day=27): Holiday.spring_festival.value,
        datetime.date(year=2017, month=1, day=28): Holiday.spring_festival.value,
        datetime.date(year=2017, month=1, day=29): Holiday.spring_festival.value,
        datetime.date(year=2017, month=1, day=30): Holiday.spring_festival.value,
        datetime.date(year=2017, month=1, day=31): Holiday.spring_festival.value,
        datetime.date(year=2017, month=2, day=1): Holiday.spring_festival.value,
        datetime.date(year=2017, month=2, day=2): Holiday.spring_festival.value,
        datetime.date(year=2017, month=4, day=2): Holiday.tomb_sweeping_day.value,
        datetime.date(year=2017, month=4, day=3): Holiday.tomb_sweeping_day.value,
        datetime.date(year=2017, month=4, day=4): Holiday.tomb_sweeping_day.value,
        datetime.date(year=2017, month=5, day=1): Holiday.labour_day.value,
        datetime.date(year=2017, month=5, day=28): Holiday.dragon_boat_festival.value,
        datetime.date(year=2017, month=5, day=29): Holiday.dragon_boat_festival.value,
        datetime.date(year=2017, month=5, day=30): Holiday.dragon_boat_festival.value,
        datetime.date(year=2017, month=10, day=1): Holiday.national_day.value,
        datetime.date(year=2017, month=10, day=2): Holiday.national_day.value,
        datetime.date(year=2017, month=10, day=3): Holiday.national_day.value,
        datetime.date(year=2017, month=10, day=4): Holiday.mid_autumn_festival.value,
        datetime.date(year=2017, month=10, day=5): Holiday.national_day.value,
        datetime.date(year=2017, month=10, day=6): Holiday.national_day.value,
        datetime.date(year=2017, month=10, day=7): Holiday.national_day.value,
        datetime.date(year=2017, month=10, day=8): Holiday.national_day.value,
        datetime.date(year=2018, month=1, day=1): Holiday.new_years_day.value,
        datetime.date(year=2018, month=2, day=15): Holiday.spring_festival.value,
        datetime.date(year=2018, month=2, day=16): Holiday.spring_festival.value,
        datetime.date(year=2018, month=2, day=17): Holiday.spring_festival.value,
        datetime.date(year=2018, month=2, day=18): Holiday.spring_festival.value,
        datetime.date(year=2018, month=2, day=19): Holiday.spring_festival.value,
        datetime.date(year=2018, month=2, day=20): Holiday.spring_festival.value,
        datetime.date(year=2018, month=2, day=21): Holiday.spring_festival.value,
        datetime.date(year=2018, month=4, day=5): Holiday.tomb_sweeping_day.value,
        datetime.date(year=2018, month=4, day=6): Holiday.tomb_sweeping_day.value,
        datetime.date(year=2018, month=4, day=7): Holiday.tomb_sweeping_day.value,
        datetime.date(year=2018, month=4, day=29): Holiday.labour_day.value,
        datetime.date(year=2018, month=4, day=30): Holiday.labour_day.value,
        datetime.date(year=2018, month=5, day=1): Holiday.labour_day.value,
        datetime.date(year=2018, month=6, day=18): Holiday.dragon_boat_festival.value,
        datetime.date(year=2018, month=9, day=24): Holiday.mid_autumn_festival.value,
        datetime.date(year=2018, month=10, day=1): Holiday.national_day.value,
        datetime.date(year=2018, month=10, day=2): Holiday.national_day.value,
        datetime.date(year=2018, month=10, day=3): Holiday.national_day.value,
        datetime.date(year=2018, month=10, day=4): Holiday.national_day.value,
        datetime.date(year=2018, month=10, day=5): Holiday.national_day.value,
        datetime.date(year=2018, month=10, day=6): Holiday.national_day.value,
        datetime.date(year=2018, month=10, day=7): Holiday.national_day.value,
        datetime.date(year=2018, month=12, day=30): Holiday.new_years_day.value,
        datetime.date(year=2018, month=12, day=31): Holiday.new_years_day.value,
        datetime.date(year=2019, month=1, day=1): Holiday.new_years_day.value,
        datetime.date(year=2019, month=2, day=4): Holiday.spring_festival.value,
        datetime.date(year=2019, month=2, day=5): Holiday.spring_festival.value,
        datetime.date(year=2019, month=2, day=6): Holiday.spring_festival.value,
        datetime.date(year=2019, month=2, day=7): Holiday.spring_festival.value,
        datetime.date(year=2019, month=2, day=8): Holiday.spring_festival.value,
        datetime.date(year=2019, month=2, day=9): Holiday.spring_festival.value,
        datetime.date(year=2019, month=2, day=10): Holiday.spring_festival.value,
        datetime.date(year=2019, month=4, day=5): Holiday.tomb_sweeping_day.value,
        datetime.date(year=2019, month=4, day=6): Holiday.tomb_sweeping_day.value,
        datetime.date(year=2019, month=4, day=7): Holiday.tomb_sweeping_day.value,
        datetime.date(year=2019, month=5, day=1): Holiday.labour_day.value,
        datetime.date(year=2019, month=5, day=2): Holiday.labour_day.value,
        datetime.date(year=2019, month=5, day=3): Holiday.labour_day.value,
        datetime.date(year=2019, month=5, day=4): Holiday.labour_day.value,
        datetime.date(year=2019, month=6, day=7): Holiday.dragon_boat_festival.value,
        datetime.date(year=2019, month=6, day=8): Holiday.dragon_boat_festival.value,
        datetime.date(year=2019, month=6, day=9): Holiday.dragon_boat_festival.value,
        datetime.date(year=2019, month=9, day=13): Holiday.mid_autumn_festival.value,
        datetime.date(year=2019, month=9, day=14): Holiday.mid_autumn_festival.value,
        datetime.date(year=2019, month=9, day=15): Holiday.mid_autumn_festival.value,
        datetime.date(year=2019, month=10, day=1): Holiday.national_day.value,
        datetime.date(year=2019, month=10, day=2): Holiday.national_day.value,
        datetime.date(year=2019, month=10, day=3): Holiday.national_day.value,
        datetime.date(year=2019, month=10, day=4): Holiday.national_day.value,
        datetime.date(year=2019, month=10, day=5): Holiday.national_day.value,
        datetime.date(year=2019, month=10, day=6): Holiday.national_day.value,
        datetime.date(year=2019, month=10, day=7): Holiday.national_day.value,
        datetime.date(year=2020, month=1, day=1): Holiday.new_years_day.value,
        datetime.date(year=2020, month=1, day=24): Holiday.spring_festival.value,
        datetime.date(year=2020, month=1, day=25): Holiday.spring_festival.value,
        datetime.date(year=2020, month=1, day=26): Holiday.spring_festival.value,
        datetime.date(year=2020, month=1, day=27): Holiday.spring_festival.value,
        datetime.date(year=2020, month=1, day=28): Holiday.spring_festival.value,
        datetime.date(year=2020, month=1, day=29): Holiday.spring_festival.value,
        datetime.date(year=2020, month=1, day=30): Holiday.spring_festival.value,
        datetime.date(year=2020, month=1, day=31): Holiday.spring_festival.value,
        datetime.date(year=2020, month=2, day=1): Holiday.spring_festival.value,
        datetime.date(year=2020, month=2, day=2): Holiday.spring_festival.value,
        datetime.date(year=2020, month=4, day=4): Holiday.tomb_sweeping_day.value,
        datetime.date(year=2020, month=4, day=5): Holiday.tomb_sweeping_day.value,
        datetime.date(year=2020, month=4, day=6): Holiday.tomb_sweeping_day.value,
        datetime.date(year=2020, month=5, day=1): Holiday.labour_day.value,
        datetime.date(year=2020, month=5, day=2): Holiday.labour_day.value,
        datetime.date(year=2020, month=5, day=3): Holiday.labour_day.value,
        datetime.date(year=2020, month=5, day=4): Holiday.labour_day.value,
        datetime.date(year=2020, month=5, day=5): Holiday.labour_day.value,
        datetime.date(year=2020, month=6, day=25): Holiday.dragon_boat_festival.value,
        datetime.date(year=2020, month=6, day=26): Holiday.dragon_boat_festival.value,
        datetime.date(year=2020, month=6, day=27): Holiday.dragon_boat_festival.value,
        datetime.date(year=2020, month=10, day=1): Holiday.national_day.value,
        datetime.date(year=2020, month=10, day=2): Holiday.national_day.value,
        datetime.date(year=2020, month=10, day=3): Holiday.national_day.value,
        datetime.date(year=2020, month=10, day=4): Holiday.national_day.value,
        datetime.date(year=2020, month=10, day=5): Holiday.national_day.value,
        datetime.date(year=2020, month=10, day=6): Holiday.national_day.value,
        datetime.date(year=2020, month=10, day=7): Holiday.national_day.value,
        datetime.date(year=2020, month=10, day=8): Holiday.national_day.value,
        datetime.date(year=2021, month=1, day=1): Holiday.new_years_day.value,
        datetime.date(year=2021, month=1, day=2): Holiday.new_years_day.value,
        datetime.date(year=2021, month=1, day=3): Holiday.new_years_day.value,
        datetime.date(year=2021, month=2, day=11): Holiday.spring_festival.value,
        datetime.date(year=2021, month=2, day=12): Holiday.spring_festival.value,
        datetime.date(year=2021, month=2, day=13): Holiday.spring_festival.value,
        datetime.date(year=2021, month=2, day=14): Holiday.spring_festival.value,
        datetime.date(year=2021, month=2, day=15): Holiday.spring_festival.value,
        datetime.date(year=2021, month=2, day=16): Holiday.spring_festival.value,
        datetime.date(year=2021, month=2, day=17): Holiday.spring_festival.value,
        datetime.date(year=2021, month=4, day=3): Holiday.tomb_sweeping_day.value,
        datetime.date(year=2021, month=4, day=4): Holiday.tomb_sweeping_day.value,
        datetime.date(year=2021, month=4, day=5): Holiday.tomb_sweeping_day.value,
        datetime.date(year=2021, month=5, day=1): Holiday.labour_day.value,
        datetime.date(year=2021, month=5, day=2): Holiday.labour_day.value,
        datetime.date(year=2021, month=5, day=3): Holiday.labour_day.value,
        datetime.date(year=2021, month=5, day=4): Holiday.labour_day.value,
        datetime.date(year=2021, month=5, day=5): Holiday.labour_day.value,
        datetime.date(year=2021, month=6, day=12): Holiday.dragon_boat_festival.value,
        datetime.date(year=2021, month=6, day=13): Holiday.dragon_boat_festival.value,
        datetime.date(year=2021, month=6, day=14): Holiday.dragon_boat_festival.value,
        datetime.date(year=2021, month=9, day=19): Holiday.mid_autumn_festival.value,
        datetime.date(year=2021, month=9, day=20): Holiday.mid_autumn_festival.value,
        datetime.date(year=2021, month=9, day=21): Holiday.mid_autumn_festival.value,
        datetime.date(year=2021, month=10, day=1): Holiday.national_day.value,
        datetime.date(year=2021, month=10, day=2): Holiday.national_day.value,
        datetime.date(year=2021, month=10, day=3): Holiday.national_day.value,
        datetime.date(year=2021, month=10, day=4): Holiday.national_day.value,
        datetime.date(year=2021, month=10, day=5): Holiday.national_day.value,
        datetime.date(year=2021, month=10, day=6): Holiday.national_day.value,
        datetime.date(year=2021, month=10, day=7): Holiday.national_day.value,
        datetime.date(year=2022, month=1, day=1): Holiday.new_years_day.value,
        datetime.date(year=2022, month=1, day=2): Holiday.new_years_day.value,
        datetime.date(year=2022, month=1, day=3): Holiday.new_years_day.value,
        datetime.date(year=2022, month=1, day=31): Holiday.spring_festival.value,
        datetime.date(year=2022, month=2, day=1): Holiday.spring_festival.value,
        datetime.date(year=2022, month=2, day=2): Holiday.spring_festival.value,
        datetime.date(year=2022, month=2, day=3): Holiday.spring_festival.value,
        datetime.date(year=2022, month=2, day=4): Holiday.spring_festival.value,
        datetime.date(year=2022, month=2, day=5): Holiday.spring_festival.value,
        datetime.date(year=2022, month=2, day=6): Holiday.spring_festival.value,
        datetime.date(year=2022, month=4, day=3): Holiday.tomb_sweeping_day.value,
        datetime.date(year=2022, month=4, day=4): Holiday.tomb_sweeping_day.value,
        datetime.date(year=2022, month=4, day=5): Holiday.tomb_sweeping_day.value,
        datetime.date(year=2022, month=4, day=30): Holiday.labour_day.value,
        datetime.date(year=2022, month=5, day=1): Holiday.labour_day.value,
        datetime.date(year=2022, month=5, day=2): Holiday.labour_day.value,
        datetime.date(year=2022, month=5, day=3): Holiday.labour_day.value,
        datetime.date(year=2022, month=5, day=4): Holiday.labour_day.value,
        datetime.date(year=2022, month=6, day=3): Holiday.dragon_boat_festival.value,
        datetime.date(year=2022, month=6, day=4): Holiday.dragon_boat_festival.value,
        datetime.date(year=2022, month=6, day=5): Holiday.dragon_boat_festival.value,
        datetime.date(year=2022, month=9, day=10): Holiday.mid_autumn_festival.value,
        datetime.date(year=2022, month=9, day=11): Holiday.mid_autumn_festival.value,
        datetime.date(year=2022, month=9, day=12): Holiday.mid_autumn_festival.value,
        datetime.date(year=2022, month=10, day=1): Holiday.national_day.value,
        datetime.date(year=2022, month=10, day=2): Holiday.national_day.value,
        datetime.date(year=2022, month=10, day=3): Holiday.national_day.value,
        datetime.date(year=2022, month=10, day=4): Holiday.national_day.value,
        datetime.date(year=2022, month=10, day=5): Holiday.national_day.value,
        datetime.date(year=2022, month=10, day=6): Holiday.national_day.value,
        datetime.date(year=2022, month=10, day=7): Holiday.national_day.value,
        datetime.date(year=2022, month=12, day=31): Holiday.new_years_day.value,
        datetime.date(year=2023, month=1, day=1): Holiday.new_years_day.value,
        datetime.date(year=2023, month=1, day=2): Holiday.new_years_day.value,
        datetime.date(year=2023, month=1, day=21): Holiday.spring_festival.value,
        datetime.date(year=2023, month=1, day=22): Holiday.spring_festival.value,
        datetime.date(year=2023, month=1, day=23): Holiday.spring_festival.value,
        datetime.date(year=2023, month=1, day=24): Holiday.spring_festival.value,
        datetime.date(year=2023, month=1, day=25): Holiday.spring_festival.value,
        datetime.date(year=2023, month=1, day=26): Holiday.spring_festival.value,
        datetime.date(year=2023, month=1, day=27): Holiday.spring_festival.value,
        datetime.date(year=2023, month=4, day=5): Holiday.tomb_sweeping_day.value,
        datetime.date(year=2023, month=4, day=29): Holiday.labour_day.value,
        datetime.date(year=2023, month=4, day=30): Holiday.labour_day.value,
        datetime.date(year=2023, month=5, day=1): Holiday.labour_day.value,
        datetime.date(year=2023, month=5, day=2): Holiday.labour_day.value,
        datetime.date(year=2023, month=5, day=3): Holiday.labour_day.value,
        datetime.date(year=2023, month=6, day=22): Holiday.dragon_boat_festival.value,
        datetime.date(year=2023, month=6, day=23): Holiday.dragon_boat_festival.value,
        datetime.date(year=2023, month=6, day=24): Holiday.dragon_boat_festival.value,
        datetime.date(year=2023, month=9, day=29): Holiday.mid_autumn_festival.value,
        datetime.date(year=2023, month=9, day=30): Holiday.national_day.value,
        datetime.date(year=2023, month=10, day=1): Holiday.national_day.value,
        datetime.date(year=2023, month=10, day=2): Holiday.national_day.value,
        datetime.date(year=2023, month=10, day=3): Holiday.national_day.value,
        datetime.date(year=2023, month=10, day=4): Holiday.national_day.value,
        datetime.date(year=2023, month=10, day=5): Holiday.national_day.value,
        datetime.date(year=2023, month=10, day=6): Holiday.national_day.value,
        datetime.date(year=2024, month=1, day=1): Holiday.new_years_day.value,
        datetime.date(year=2024, month=2, day=9): Holiday.spring_festival.value,  # 除夕上班但是不开盘
        datetime.date(year=2024, month=2, day=10): Holiday.spring_festival.value,
        datetime.date(year=2024, month=2, day=11): Holiday.spring_festival.value,
        datetime.date(year=2024, month=2, day=12): Holiday.spring_festival.value,
        datetime.date(year=2024, month=2, day=13): Holiday.spring_festival.value,
        datetime.date(year=2024, month=2, day=14): Holiday.spring_festival.value,
        datetime.date(year=2024, month=2, day=15): Holiday.spring_festival.value,
        datetime.date(year=2024, month=2, day=16): Holiday.spring_festival.value,
        datetime.date(year=2024, month=2, day=17): Holiday.spring_festival.value,
        datetime.date(year=2024, month=4, day=4): Holiday.tomb_sweeping_day.value,
        datetime.date(year=2024, month=4, day=5): Holiday.tomb_sweeping_day.value,
        datetime.date(year=2024, month=4, day=6): Holiday.tomb_sweeping_day.value,
        datetime.date(year=2024, month=5, day=1): Holiday.labour_day.value,
        datetime.date(year=2024, month=5, day=2): Holiday.labour_day.value,
        datetime.date(year=2024, month=5, day=3): Holiday.labour_day.value,
        datetime.date(year=2024, month=5, day=4): Holiday.labour_day.value,
        datetime.date(year=2024, month=5, day=5): Holiday.labour_day.value,
        datetime.date(year=2024, month=6, day=10): Holiday.dragon_boat_festival.value,
        datetime.date(year=2024, month=9, day=15): Holiday.mid_autumn_festival.value,
        datetime.date(year=2024, month=9, day=16): Holiday.mid_autumn_festival.value,
        datetime.date(year=2024, month=9, day=17): Holiday.mid_autumn_festival.value,
        datetime.date(year=2024, month=10, day=1): Holiday.national_day.value,
        datetime.date(year=2024, month=10, day=2): Holiday.national_day.value,
        datetime.date(year=2024, month=10, day=3): Holiday.national_day.value,
        datetime.date(year=2024, month=10, day=4): Holiday.national_day.value,
        datetime.date(year=2024, month=10, day=5): Holiday.national_day.value,
        datetime.date(year=2024, month=10, day=6): Holiday.national_day.value,
        datetime.date(year=2024, month=10, day=7): Holiday.national_day.value,
        datetime.date(year=2025, month=1, day=1): Holiday.new_years_day.value,
        datetime.date(year=2025, month=1, day=28): Holiday.spring_festival.value,
        datetime.date(year=2025, month=1, day=29): Holiday.spring_festival.value,
        datetime.date(year=2025, month=1, day=30): Holiday.spring_festival.value,
        datetime.date(year=2025, month=1, day=31): Holiday.spring_festival.value,
        datetime.date(year=2025, month=2, day=1): Holiday.spring_festival.value,
        datetime.date(year=2025, month=2, day=2): Holiday.spring_festival.value,
        datetime.date(year=2025, month=2, day=3): Holiday.spring_festival.value,
        datetime.date(year=2025, month=2, day=4): Holiday.spring_festival.value,
        datetime.date(year=2025, month=4, day=4): Holiday.tomb_sweeping_day.value,
        datetime.date(year=2025, month=4, day=5): Holiday.tomb_sweeping_day.value,
        datetime.date(year=2025, month=4, day=6): Holiday.tomb_sweeping_day.value,
        datetime.date(year=2025, month=5, day=1): Holiday.labour_day.value,
        datetime.date(year=2025, month=5, day=2): Holiday.labour_day.value,
        datetime.date(year=2025, month=5, day=3): Holiday.labour_day.value,
        datetime.date(year=2025, month=5, day=4): Holiday.labour_day.value,
        datetime.date(year=2025, month=5, day=5): Holiday.labour_day.value,
        datetime.date(year=2025, month=5, day=31): Holiday.dragon_boat_festival.value,
        datetime.date(year=2025, month=6, day=1): Holiday.dragon_boat_festival.value,
        datetime.date(year=2025, month=6, day=2): Holiday.dragon_boat_festival.value,
        datetime.date(year=2025, month=10, day=1): Holiday.national_day.value,
        datetime.date(year=2025, month=10, day=2): Holiday.national_day.value,
        datetime.date(year=2025, month=10, day=3): Holiday.national_day.value,
        datetime.date(year=2025, month=10, day=4): Holiday.national_day.value,
        datetime.date(year=2025, month=10, day=5): Holiday.national_day.value,
        datetime.date(year=2025, month=10, day=6): Holiday.mid_autumn_festival.value,
        datetime.date(year=2025, month=10, day=7): Holiday.national_day.value,
        datetime.date(year=2025, month=10, day=8): Holiday.national_day.value,
        # 上海票据交易所关于中国票据业务系统2025年节假日调整和2026年节假日预设的通知：
        # https://mp.weixin.qq.com/s/9eHii0bu6NRKtekeRwCl2Q
        # todo: 2026年节假日已经根据国务院正式通知更新, 与中国票据业务系统预设一致
        datetime.date(year=2026, month=1, day=1): Holiday.new_years_day.value,
        datetime.date(year=2026, month=1, day=2): Holiday.new_years_day.value,
        datetime.date(year=2026, month=1, day=3): Holiday.new_years_day.value,
        datetime.date(year=2026, month=2, day=15): Holiday.spring_festival.value,
        datetime.date(year=2026, month=2, day=16): Holiday.spring_festival.value,
        datetime.date(year=2026, month=2, day=17): Holiday.spring_festival.value,
        datetime.date(year=2026, month=2, day=18): Holiday.spring_festival.value,
        datetime.date(year=2026, month=2, day=19): Holiday.spring_festival.value,
        datetime.date(year=2026, month=2, day=20): Holiday.spring_festival.value,
        datetime.date(year=2026, month=2, day=21): Holiday.spring_festival.value,
        datetime.date(year=2026, month=2, day=22): Holiday.spring_festival.value,
        datetime.date(year=2026, month=2, day=23): Holiday.spring_festival.value,
        datetime.date(year=2026, month=4, day=4): Holiday.tomb_sweeping_day.value,
        datetime.date(year=2026, month=4, day=5): Holiday.tomb_sweeping_day.value,
        datetime.date(year=2026, month=4, day=6): Holiday.tomb_sweeping_day.value,
        datetime.date(year=2026, month=5, day=1): Holiday.labour_day.value,
        datetime.date(year=2026, month=5, day=2): Holiday.labour_day.value,
        datetime.date(year=2026, month=5, day=3): Holiday.labour_day.value,
        datetime.date(year=2026, month=5, day=4): Holiday.labour_day.value,
        datetime.date(year=2026, month=5, day=5): Holiday.labour_day.value,
        datetime.date(year=2026, month=6, day=19): Holiday.dragon_boat_festival.value,
        datetime.date(year=2026, month=6, day=20): Holiday.dragon_boat_festival.value,
        datetime.date(year=2026, month=6, day=21): Holiday.dragon_boat_festival.value,
        datetime.date(year=2026, month=9, day=25): Holiday.mid_autumn_festival.value,
        datetime.date(year=2026, month=9, day=26): Holiday.mid_autumn_festival.value,
        datetime.date(year=2026, month=9, day=27): Holiday.mid_autumn_festival.value,
        datetime.date(year=2026, month=10, day=1): Holiday.national_day.value,
        datetime.date(year=2026, month=10, day=2): Holiday.national_day.value,
        datetime.date(year=2026, month=10, day=3): Holiday.national_day.value,
        datetime.date(year=2026, month=10, day=4): Holiday.national_day.value,
        datetime.date(year=2026, month=10, day=5): Holiday.national_day.value,
        datetime.date(year=2026, month=10, day=6): Holiday.national_day.value,
        datetime.date(year=2026, month=10, day=7): Holiday.national_day.value,
        # todo: 2027年节假日根据农历预测，具体以国务院通知为准
        datetime.date(year=2027, month=1, day=1): Holiday.new_years_day.value,
        datetime.date(year=2027, month=2, day=5): Holiday.spring_festival.value,
        datetime.date(year=2027, month=2, day=6): Holiday.spring_festival.value,
        datetime.date(year=2027, month=2, day=7): Holiday.spring_festival.value,
        datetime.date(year=2027, month=2, day=8): Holiday.spring_festival.value,
        datetime.date(year=2027, month=2, day=9): Holiday.spring_festival.value,
        datetime.date(year=2027, month=2, day=10): Holiday.spring_festival.value,
        datetime.date(year=2027, month=2, day=11): Holiday.spring_festival.value,
        datetime.date(year=2027, month=2, day=12): Holiday.spring_festival.value,
        datetime.date(year=2027, month=4, day=3): Holiday.tomb_sweeping_day.value,
        datetime.date(year=2027, month=4, day=4): Holiday.tomb_sweeping_day.value,
        datetime.date(year=2027, month=4, day=5): Holiday.tomb_sweeping_day.value,
        datetime.date(year=2027, month=5, day=1): Holiday.labour_day.value,
        datetime.date(year=2027, month=5, day=2): Holiday.labour_day.value,
        datetime.date(year=2027, month=5, day=3): Holiday.labour_day.value,
        datetime.date(year=2027, month=5, day=4): Holiday.labour_day.value,
        datetime.date(year=2027, month=5, day=5): Holiday.labour_day.value,
        datetime.date(year=2027, month=6, day=9): Holiday.dragon_boat_festival.value,
        datetime.date(year=2027, month=6, day=10): Holiday.dragon_boat_festival.value,
        datetime.date(year=2027, month=6, day=11): Holiday.dragon_boat_festival.value,
        datetime.date(year=2027, month=9, day=15): Holiday.mid_autumn_festival.value,
        datetime.date(year=2027, month=10, day=1): Holiday.national_day.value,
        datetime.date(year=2027, month=10, day=2): Holiday.national_day.value,
        datetime.date(year=2027, month=10, day=3): Holiday.national_day.value,
        datetime.date(year=2027, month=10, day=4): Holiday.national_day.value,
        datetime.date(year=2027, month=10, day=5): Holiday.national_day.value,
        datetime.date(year=2027, month=10, day=6): Holiday.national_day.value,
        datetime.date(year=2027, month=10, day=7): Holiday.national_day.value,
        # todo: 2028年节假日根据农历预测，具体以国务院通知为准
        datetime.date(year=2028, month=1, day=1): Holiday.new_years_day.value,
        datetime.date(year=2028, month=1, day=2): Holiday.new_years_day.value,
        datetime.date(year=2028, month=1, day=3): Holiday.new_years_day.value,
        datetime.date(year=2028, month=1, day=25): Holiday.spring_festival.value,
        datetime.date(year=2028, month=1, day=26): Holiday.spring_festival.value,
        datetime.date(year=2028, month=1, day=27): Holiday.spring_festival.value,
        datetime.date(year=2028, month=1, day=28): Holiday.spring_festival.value,
        datetime.date(year=2028, month=1, day=29): Holiday.spring_festival.value,
        datetime.date(year=2028, month=1, day=30): Holiday.spring_festival.value,
        datetime.date(year=2028, month=1, day=31): Holiday.spring_festival.value,
        datetime.date(year=2028, month=2, day=1): Holiday.spring_festival.value,
        datetime.date(year=2028, month=4, day=2): Holiday.tomb_sweeping_day.value,
        datetime.date(year=2028, month=4, day=3): Holiday.tomb_sweeping_day.value,
        datetime.date(year=2028, month=4, day=4): Holiday.tomb_sweeping_day.value,
        datetime.date(year=2028, month=4, day=29): Holiday.labour_day.value,
        datetime.date(year=2028, month=4, day=30): Holiday.labour_day.value,
        datetime.date(year=2028, month=5, day=1): Holiday.labour_day.value,
        datetime.date(year=2028, month=5, day=2): Holiday.labour_day.value,
        datetime.date(year=2028, month=5, day=3): Holiday.labour_day.value,
        datetime.date(year=2028, month=6, day=25): Holiday.dragon_boat_festival.value,
        datetime.date(year=2028, month=6, day=26): Holiday.dragon_boat_festival.value,
        datetime.date(year=2028, month=6, day=27): Holiday.dragon_boat_festival.value,
        datetime.date(year=2028, month=10, day=1): Holiday.national_day.value,
        datetime.date(year=2028, month=10, day=2): Holiday.national_day.value,
        datetime.date(year=2028, month=10, day=3): Holiday.mid_autumn_festival.value,
        datetime.date(year=2028, month=10, day=4): Holiday.national_day.value,
        datetime.date(year=2028, month=10, day=5): Holiday.national_day.value,
        datetime.date(year=2028, month=10, day=6): Holiday.national_day.value,
        datetime.date(year=2028, month=10, day=7): Holiday.national_day.value,
        datetime.date(year=2028, month=10, day=8): Holiday.national_day.value,
        # todo: 2029年节假日根据农历预测，具体以国务院通知为准
        datetime.date(year=2028, month=12, day=31): Holiday.new_years_day.value,
        datetime.date(year=2029, month=1, day=1): Holiday.new_years_day.value,
        datetime.date(year=2029, month=1, day=2): Holiday.new_years_day.value,
        datetime.date(year=2029, month=2, day=12): Holiday.spring_festival.value,
        datetime.date(year=2029, month=2, day=13): Holiday.spring_festival.value,
        datetime.date(year=2029, month=2, day=14): Holiday.spring_festival.value,
        datetime.date(year=2029, month=2, day=15): Holiday.spring_festival.value,
        datetime.date(year=2029, month=2, day=16): Holiday.spring_festival.value,
        datetime.date(year=2029, month=2, day=17): Holiday.spring_festival.value,
        datetime.date(year=2029, month=2, day=18): Holiday.spring_festival.value,
        datetime.date(year=2029, month=2, day=19): Holiday.spring_festival.value,
        datetime.date(year=2029, month=4, day=4): Holiday.tomb_sweeping_day.value,
        datetime.date(year=2029, month=4, day=28): Holiday.labour_day.value,
        datetime.date(year=2029, month=4, day=29): Holiday.labour_day.value,
        datetime.date(year=2029, month=4, day=30): Holiday.labour_day.value,
        datetime.date(year=2029, month=5, day=1): Holiday.labour_day.value,
        datetime.date(year=2029, month=5, day=2): Holiday.labour_day.value,
        datetime.date(year=2029, month=6, day=16): Holiday.dragon_boat_festival.value,
        datetime.date(year=2029, month=6, day=17): Holiday.dragon_boat_festival.value,
        datetime.date(year=2029, month=6, day=18): Holiday.dragon_boat_festival.value,
        datetime.date(year=2029, month=9, day=21): Holiday.mid_autumn_festival.value,
        datetime.date(year=2029, month=9, day=22): Holiday.mid_autumn_festival.value,
        datetime.date(year=2029, month=9, day=23): Holiday.mid_autumn_festival.value,
        datetime.date(year=2029, month=10, day=1): Holiday.national_day.value,
        datetime.date(year=2029, month=10, day=2): Holiday.national_day.value,
        datetime.date(year=2029, month=10, day=3): Holiday.mid_autumn_festival.value,
        datetime.date(year=2029, month=10, day=4): Holiday.national_day.value,
        datetime.date(year=2029, month=10, day=5): Holiday.national_day.value,
        datetime.date(year=2029, month=10, day=6): Holiday.national_day.value,
        datetime.date(year=2029, month=10, day=7): Holiday.national_day.value,
        # todo: 2030年节假日根据农历预测，具体以国务院通知为准
        datetime.date(year=2029, month=12, day=30): Holiday.new_years_day.value,
        datetime.date(year=2029, month=12, day=31): Holiday.new_years_day.value,
        datetime.date(year=2030, month=1, day=1): Holiday.new_years_day.value,
        datetime.date(year=2030, month=2, day=2): Holiday.spring_festival.value,
        datetime.date(year=2030, month=2, day=3): Holiday.spring_festival.value,
        datetime.date(year=2030, month=2, day=4): Holiday.spring_festival.value,
        datetime.date(year=2030, month=2, day=5): Holiday.spring_festival.value,
        datetime.date(year=2030, month=2, day=6): Holiday.spring_festival.value,
        datetime.date(year=2030, month=2, day=7): Holiday.spring_festival.value,
        datetime.date(year=2030, month=2, day=8): Holiday.spring_festival.value,
        datetime.date(year=2030, month=2, day=9): Holiday.spring_festival.value,
        datetime.date(year=2030, month=4, day=5): Holiday.tomb_sweeping_day.value,
        datetime.date(year=2030, month=4, day=6): Holiday.tomb_sweeping_day.value,
        datetime.date(year=2030, month=4, day=7): Holiday.tomb_sweeping_day.value,
        datetime.date(year=2030, month=5, day=1): Holiday.labour_day.value,
        datetime.date(year=2030, month=5, day=2): Holiday.labour_day.value,
        datetime.date(year=2030, month=5, day=3): Holiday.labour_day.value,
        datetime.date(year=2030, month=5, day=4): Holiday.labour_day.value,
        datetime.date(year=2030, month=5, day=5): Holiday.labour_day.value,
        datetime.date(year=2030, month=6, day=5): Holiday.dragon_boat_festival.value,
        datetime.date(year=2030, month=9, day=12): Holiday.mid_autumn_festival.value,
        datetime.date(year=2030, month=9, day=13): Holiday.mid_autumn_festival.value,
        datetime.date(year=2030, month=9, day=14): Holiday.mid_autumn_festival.value,
        datetime.date(year=2030, month=10, day=1): Holiday.national_day.value,
        datetime.date(year=2030, month=10, day=2): Holiday.national_day.value,
        datetime.date(year=2030, month=10, day=3): Holiday.national_day.value,
        datetime.date(year=2030, month=10, day=4): Holiday.national_day.value,
        datetime.date(year=2030, month=10, day=5): Holiday.national_day.value,
        datetime.date(year=2030, month=10, day=6): Holiday.national_day.value,
        datetime.date(year=2030, month=10, day=7): Holiday.national_day.value,
    }

    def __init__(self):
        super().__init__()
        self.holidays = set(self.holidays_dict.keys())


# 单例
CN_CALENDAR = ChineseCalendar()



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import numpy as np
from pricelib.common.pricing_engine_base import McEngine
from pricelib.common.date_handler import global_evaluation_date


class MCAirbagEngine(McEngine):
    """安全气囊 Monte Carlo 模拟定价引擎
    安全气囊是美式观察向下敲入期权，到期支付
    只支持离散观察(默认为每日观察)
    """

    def calc_present_value(self, prod, t=None, spot=None):
        """计算现值
        Args:
            prod: Product产品对象
            t: datetime.date，估值日; 如果是None，则使用全局估值日globalEvaluationDate
            spot: float，估值日标的价格，如果是None，则使用随机过程的当前价格
        Returns: float，现值；
                如果输入spot = np.ndarray一维价格向量，t=终止日，则返回盈亏测算向量，即每个价格对应的现值，而不是平均值
        """
        calculate_date = global_evaluation_date() if t is None else t
        _maturity = (prod.end_date - calculate_date).days / prod.annual_days.value
        _maturity_business_days = prod.trade_calendar.business_days_between(calculate_date, prod.end_date)
        _tenor = (prod.end_date - prod.start_date).days / prod.annual_days.value  # 从起始日到到期日的自然日时间（年）
        if spot is None:
            spot = self.process.spot()
        # 当估值日就是到期日时，直接用spot计算payoff，无需随机数矩阵
        if calculate_date == prod.end_date:
            if isinstance(spot, (list, np.ndarray)):  # spot是一个np.ndarray一维价格向量，返回盈亏测算向量，即每个价格对应的现值，而不是平均值
                paths = np.array([spot, ])
            else:  # spot是一个数值，计算期末payoff
                paths = np.array([[spot, ]])
        else:
            if spot != self.process.spot():
                self.reset_paths_flag()  # 重置路径标志位，重新生成路径
            paths = self.path_generator(n_step=_maturity_business_days, spot=spot,
                                    t_step_per_year=prod.t_step_per_year).copy()

        if prod.discrete_obs_interval is None:  # 每日观察
            obs_points = np.arange(0, _maturity_business_days + 1, 1)
        else:  # 均匀离散观察
            dt_step = prod.discrete_obs_interval * prod.t_step_per_year
            obs_points = np.flip(np.round(np.arange(_maturity_business_days, 0, -dt_step)).astype(int))
            obs_points = np.concatenate((np.array([0]), obs_points))
            obs_points[obs_points > _maturity_business_days] = _maturity_business_days  # 防止闰年导致的下标越界

        # 未敲入，期末价值是call或牛市价差
        if prod.init_strike_cap is None:
            init_payoff = paths[-1] - prod.strike
        else:  # 若存在敲入前封顶价格
            init_payoff = np.where(paths[-1] - prod.init_strike_cap > 0,
                                   prod.init_strike_cap - prod.strike, paths[-1] - prod.strike)
        # 已敲入，且标的上涨
        if prod.in_strike_cap is None:
            in_payoff_upper = paths[-1] - prod.s0
        else:  # 若存在敲入后封顶价格
            in_payoff_upper = np.where(paths[-1] - prod.in_strike_cap > 0,
                                       prod.in_strike_cap - prod.s0, paths[-1] - prod.s0)
        # 已敲入，且标的下跌
        if prod.in_strike_floor is None:
            in_payoff_lower = paths[-1] - prod.s0
        else:  # 若存在敲入后封底价格
            in_payoff_lower = np.where(paths[-1] - prod.in_strike_floor < 0,
                                       prod.in_strike_floor - prod.s0, paths[-1] - prod.s0)
        # 安全气囊，payoff汇总
        knock_in_bool = np.any(paths[obs_points] <= prod.barrier, axis=0)
        value = (np.where(knock_in_bool,
                                 # 已敲入，标的涨跌幅，可能存在保底或封顶执行价
                                 np.where(paths[-1] - prod.s0 < 0,
                                          # 敲入且标的下跌
                                          in_payoff_lower * prod.knockin_parti,  # 敲入下跌参与率
                                          # 敲入且标的上涨
                                          in_payoff_upper * prod.reset_call_parti),  # 敲入重置后看涨参与率
                                 # 未敲入，香草看涨或者牛市价差
                                 np.where(init_payoff < 0, 0, init_payoff) * prod.call_parti)  # 未敲入看涨参与率
                        ) + prod.floor_yield * prod.s0 * _tenor  # 保底收益率
        value *= self.process.interest.disc_factor(_maturity)

        if (calculate_date == prod.end_date) and isinstance(spot, np.ndarray):
            return value  # 终止日, spot是价格向量, 则返回盈亏测算向量，即每个价格对应的现值，而不是平均值
        else:
            return np.mean(value)



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import pytest
import pandas as pd
from pandas.testing import assert_frame_equal
from examples.accurals import accumulator_demo, range_accural_demo


def test_accumulator_demo():
    expected_data = [['累购(Accumulator)', -0.7201659938343952], ['累沽(Decumulator)', -1.0008722924685676]]
    expected_df = pd.DataFrame(expected_data, columns=['期权类型', 'MonteCarlo'])
    res_df = accumulator_demo.run()
    assert_frame_equal(res_df, expected_df, check_dtype=True)


def test_range_accural_demo():
    res = range_accural_demo.run()
    assert res == pytest.approx({'结构': '区间累计(Range Accural)', 'MonteCarlo': 10.167894272305867}, rel=1e-10)



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import numpy as np
from .utility import logging

try:
    import plotly.graph_objects as go
    import matplotlib.pyplot as plt
    import matplotlib.lines as mlines
except ModuleNotFoundError as e:
    logging.error(e)
    logging.warning('请安装plotly和matplotlib库,\nmatplotlib>=3.5.3, <=3.7.5;\nplotly>=5.16.1, <=5.22.0')
    exit(0)


def draw_greeks_surface(greeks_name, greeks_matrix, n_smax=None, s_step=None,
                        spot_range=None, z_range=None, show_plot=True):
    """绘制greeks曲面, spot_range是价格索引的区间
    Args:
        greeks_name: str, greeks名称, 如delta, gamma, theta等
        greeks_matrix: np.ndarray, 希腊值二维数组，行代表价格，列代表时间
        n_smax: int, 例如2 价格上界是几倍S0
        s_step: int, 例如800 价格格点数
        spot_range: tuple(int)=(spot_min, spot_max), 价格索引的区间, 默认为None, 即绘制全部价格范围的图像
        z_range: tuple(float)=(z_min, z_max), 希腊值区间, 默认为None, 即绘制全部希腊值范围的图像
        show_plot: bool, 是否显示图像, 默认为True
    Returns:
        fig: plotly.graph_objects.Figure对象，希腊值三维图
    """
    price_grid = np.array(range(1, 1 + greeks_matrix.shape[0]))
    if spot_range is not None:
        spot_min, spot_max = spot_range
        price_grid = price_grid[spot_min:spot_max + 1]
        greeks_matrix = greeks_matrix[spot_min:spot_max + 1, ]
    time_grid = np.array(range(greeks_matrix.shape[1]))
    fig = go.Figure(
        data=[go.Surface(z=greeks_matrix, x=time_grid, y=price_grid * 100 * n_smax / s_step,
                         colorscale='Turbo',  # Jet,Turbo, Rainbow
                         contours={  # x, y 轴添加等高线
                         "x": {"show": True, "color": "white"},
                         "y": {"show": True, "color": "white"},
                         # "z": {"show": True, "size": 0.01, "color": "white"},
                         })])
    # 轮廓线(可选)
    fig.update_traces(contours_z={  # 轮廓设置
        "show": True,  # 开启是否显示
        "usecolormap": True,  # 颜色设置
        "highlightcolor": "red",  # 高亮 mistyrose
        "project_z": True})
    # 设置标题、图片大小
    fig.update_layout(
        title=f'<b>{greeks_name} S-t Surface</b>',
        width=800, height=900,
        autosize=True,
        scene={
            'xaxis': {'title': '<b>time grid</b>'},
            'yaxis': {'title': '<b>price grid</b>'},
            'zaxis': {'title': f"<b>{greeks_name}</b>"},
            "camera": dict(up=dict(x=0, y=0, z=1), center=dict(x=0, y=0, z=-0.02), eye=dict(x=-2, y=0, z=0)),
        },  # 相机设置:up定义了哪个方向是上,center定义相机中心位置,eye定义了相机的位置,相机将从X轴负方向看向原点，即YOZ平面在最前面。
        # template="plotly_dark",  # 深色主题
        margin={'l': 65, 'r': 50, 'b': 65, 't': 90},
        # margin=dict(r=0, l=0, t=0, b=0, pad=0)
    )
    fig.update_scenes(yaxis_autorange='reversed')  # 反转y轴(标的价格)
    if z_range is not None:
        # 设置 z 轴的显示区间
        fig.update_scenes(zaxis=dict(range=z_range))
    if show_plot:
        # 将图保存到本地
        # pyo.plot(fig, filename=f'./greeks_surface_plot.html')
        fig.show()  # 显示图片(plotly在线)
    return fig


def draw_greeks_curve(delta_matrix, gamma_matrix=None, t=0, n_smax=None, s_step=None, spot_range=None, show_plot=True):
    """绘制t时刻的greeks曲线
    Args:
        delta_matrix: np.ndarray, delta二维数组，行代表价格，列代表时间
        gamma_matrix: np.ndarray, gamma二维数组，行代表价格，列代表时间，默认为None，此时只绘制delta曲线
        t: int, 相对于起始日的时间索引，天，默认为0是期初
        n_smax: int, 例如2 价格上界是几倍S0
        s_step: int, 例如800 价格格点数
        spot_range: tuple(int)=(spot_min, spot_max), 价格索引的区间, 默认为None, 即绘制全部价格范围的图像
        show_plot: bool, 是否显示图像, 默认为True
    Returns:
        fig: plotly.graph_objects.Figure对象，希腊值曲线图
    """
    fig = go.Figure()
    price_grid = np.array(range(1, 1 + delta_matrix.shape[0]))
    if spot_range is not None:
        spot_min, spot_max = spot_range
        price_grid = price_grid[spot_min:spot_max + 1]
        delta_matrix = delta_matrix[spot_min:spot_max + 1, ]
        if gamma_matrix is not None:
            gamma_matrix = gamma_matrix[spot_min:spot_max + 1, ]

    delta_line = go.Scatter(
        x=price_grid * 100 * n_smax / s_step,
        y=delta_matrix[:, t],
        mode='lines',
        name="<b>Delta</b>"
    )
    fig.add_trace(delta_line)
    if gamma_matrix is not None:
        gamma_line = go.Scatter(
            x=price_grid * 100 * n_smax / s_step,
            y=gamma_matrix[:, t],
            mode='lines',
            name="<b>Gamma</b>"
        )
        fig.add_trace(gamma_line)
    fig.update_layout(title=f'<b>Day {t} Greeks</b>', xaxis_title='Underlying Price',
                      yaxis_title='<b>Greeks Value</b>')  # 深色主题 , template="plotly_dark",

    if show_plot:
        # pyo.plot(fig, filename=f'./greeks_curve_plot.html')
        fig.show()
    return fig


def draw_greeks_contour(greeks_matrix, s0, barrier_in, n_smax=None, s_step=None, high_value=1.2, low_value=0.01,
                        title=None):
    """绘制greeks等高线
    Args:
        greeks_matrix: np.ndarray, 希腊值二维数组，行代表价格，列代表时间
        s0: float, s0是初始价格
        barrier_in: float, barrier_in是敲入价
        n_smax: int, 例如2 价格上界是几倍S0
        s_step: int, 例如800 价格格点数
        high_value: float, greeks的上限
        low_value: float, greeks的下限
        title: str, 图表标题
    Returns: None
    """
    barrier_in_idx = barrier_in * s_step / (n_smax * s0)
    # 绘制Delta等高线图
    high_contour = plt.contour(greeks_matrix, levels=[high_value, ])
    # 清除之前的图像
    plt.clf()
    # 获取等高线的路径点
    for collection in high_contour.collections:
        paths = collection.get_paths()
        for path in paths:
            vertices = path.vertices
            # 筛选出上半边的点（未敲入）
            upper_half_vertices = vertices[vertices[:, 1] > barrier_in_idx]
            # 绘制上半边的点（未敲入）
            plt.plot(upper_half_vertices[:, 0], upper_half_vertices[:, 1], 'orange')
    plt.contour(greeks_matrix, levels=[low_value, ], colors='royalblue')
    plt.xlabel('Days')
    plt.ylabel('Price')
    if title is None:
        title = f'雪球理论Delta区间: {low_value}-{high_value}'
    plt.title(title)
    # 创建虚拟线条对象用于图例
    high_line = mlines.Line2D([], [], color='orange', label=f'Delta={high_value}')
    low_line = mlines.Line2D([], [], color='royalblue', label=f'Delta={low_value}')
    plt.legend(handles=[high_line, low_line])
    # 控制y轴范围
    plt.ylim(barrier_in_idx, barrier_in_idx * 1.7)
    yticks = np.arange(barrier_in_idx, barrier_in_idx * 1.7, n_smax * 8000 / s_step)
    plt.yticks(yticks, labels=yticks * n_smax * s0 / s_step)
    plt.rcParams['font.sans-serif'] = ['FangSong']  # 设置中文字体
    plt.show()



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
from functools import partial
import numpy as np
from pricelib.common.utilities.enums import StatusType
from pricelib.common.date_handler import CN_CALENDAR, AnnualDays
from .autocallable_base import AutocallableBase


class BothDownSnowball(AutocallableBase):
    """双降雪球结构产品类"""

    def __init__(self, s0, barrier_out_start, barrier_out_step, barrier_in, coupon_ko_start=None, coupon_ko_step=None,
                 coupon_out_stairs=None, coupon_stair_ends=None,
                 coupon_div=None, lock_term=3, maturity=None, start_date=None, end_date=None,
                 obs_dates=None, pay_dates=None, parti_in=1, margin_lvl=1, status=StatusType.NoTouch, engine=None,
                 trade_calendar=CN_CALENDAR, annual_days=AnnualDays.N365, t_step_per_year=243,
                 stoch_process=None, s=None, r=None, q=None, vol=None, engine_method=None):
        """继承自动赎回基类AutocallableBase的参数，详见AutocallableBase的__init__方法
        Args: 以下为区别于经典雪球的参数
            barrier_out_start: float, 敲出障碍价起始值，绝对值/百分比
            barrier_out_step: float, 降敲步长，绝对值/百分比
            lock_term: int，锁定期，单位为月，锁定期内不触发敲出

            降票息有两种方式，选填其中一组参数：
                1. 每月等差下降
                coupon_ko_start: float, 敲出票息起始值，百分比，年化
                coupon_ko_step: float, 降票息步长，百分比，年化
                2. 分阶段下降
                coupon_out_stairs: tuple[float], 分段票息，百分比，年化。例如(0.18, 0.05)
                coupon_stair_ends: tuple[int], 分段票息节点, 通常为(12, 24), 代表每个阶段到第几个月末为止
        """
        super().__init__(s0=s0, maturity=maturity, start_date=start_date, end_date=end_date, lock_term=lock_term,
                         trade_calendar=trade_calendar, obs_dates=obs_dates, pay_dates=pay_dates, status=status,
                         annual_days=annual_days, parti_in=parti_in, margin_lvl=margin_lvl,
                         t_step_per_year=t_step_per_year, engine=engine, stoch_process=stoch_process,
                         s=s, r=r, q=q, vol=vol, engine_method=engine_method)
        len_obs_dates = len(self.obs_dates.date_schedule)

        barrier_out = np.arange(barrier_out_start, barrier_out_start - barrier_out_step * len_obs_dates,
                                -barrier_out_step)[:len_obs_dates]
        self.barrier_out = barrier_out
        self.barrier_in = np.ones(len_obs_dates) * barrier_in

        self.coupon_ko_start = coupon_ko_start
        self.coupon_ko_step = coupon_ko_step
        self.coupon_out_stairs = coupon_out_stairs
        self.coupon_stair_ends = coupon_stair_ends
        step_true = (coupon_ko_start is not None) and (coupon_ko_step is not None)
        step_false = (coupon_ko_start is None) and (coupon_ko_step is None)
        stair_true = (coupon_out_stairs is not None) and (coupon_stair_ends is not None)
        stair_false = (coupon_out_stairs is None) and (coupon_stair_ends is None)
        if step_true and stair_false:
            self.coupon_type = "step"
            coupon_out = np.arange(coupon_ko_start, coupon_ko_start - coupon_ko_step * len_obs_dates, -coupon_ko_step)[
                         :len_obs_dates]
            self.coupon_out = coupon_out
            # 反算敲出票息起点
            self.calc_coupon = partial(self.calc_param, reset_param_method=self.reset_coupon_ko_start)
        elif stair_true and step_false:
            assert len(coupon_out_stairs) == len(coupon_stair_ends), ("输入的分段票息coupon_out_stairs"
                                                                      "与分段票息节点coupon_stair_ends的长度必须相等！")
            self.coupon_type = "stair"
            coupon_out = np.zeros(len_obs_dates + lock_term - 1)
            coupon_stair_ends = [0] + list(coupon_stair_ends)
            for i in range(len(coupon_out_stairs)):
                coupon_out[coupon_stair_ends[i]:coupon_stair_ends[i + 1]] = coupon_out_stairs[i]
            self.coupon_out = coupon_out[lock_term - 1:]
            # 反算分段敲出票息，额外参数 nth：int, 反算第几个阶段的票息
            self.calc_coupon = partial(self.calc_param, reset_param_method=self.reset_coupon_nth)
        else:
            raise ValueError("只支持两种票息输入方式:\n"
                             "1.等差下降，仅输入敲出票息起始值coupon_ko_start和降票息步长coupon_ko_step;\n"
                             "2.分阶段下降，仅输入分段票息列表coupon_out_stairs和分段票息节点列表coupon_stair_ends.")

        self.coupon_div = coupon_div if coupon_div is not None else self.coupon_out[-1]
        self.parti_out = 0
        self.strike_upper = s0
        self.strike_lower = 0

    def __repr__(self):
        """返回期权的描述"""
        return "双降雪球"

    def reset_coupon_ko_start(self, coupon_ko_start: float, coupon_div=None, *args, **kwargs):
        """重新设置敲出票息起点"""
        len_obs_dates = len(self.obs_dates.date_schedule)
        self.coupon_out = np.arange(coupon_ko_start, coupon_ko_start - self.coupon_ko_step * len_obs_dates,
                                    -self.coupon_ko_step)[:len_obs_dates]
        if coupon_div is None:  # 未指定红利票息时，将红利票息的值更新为最后一个敲出票息
            self.coupon_div = self.coupon_out[-1]

    def reset_coupon_nth(self, coupon: float, nth: int = 1, coupon_div=None, *args, **kwargs):
        """重新设置指定分段的敲出票息
        Args:
            coupon: float, 重设票息值
            coupon_div: float, 是否指定了红利票息的值
            nth: int, 修改的是第几个阶段的票息
        """
        if nth == 1:
            self.coupon_out[:self.coupon_stair_ends[0] - self.lock_term + 1] = coupon
        else:
            self.coupon_out[self.coupon_stair_ends[nth - 2] - self.lock_term + 1:
                            self.coupon_stair_ends[nth - 1] - self.lock_term + 1] = coupon
        # 如果反算的是最后一段票息，并且没有指定红利票息，同时更新红利票息的值
        if (len(self.coupon_stair_ends) == nth) and (coupon_div is None):
            self.coupon_div = coupon



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import numpy as np
from pricelib.common.utilities.enums import StatusType, ExerciseType
from pricelib.common.date_handler import CN_CALENDAR, AnnualDays
from .autocallable_base import AutocallableBase


class EuropeanSnowball(AutocallableBase):
    """欧式雪球
       保本(strike_lower=s0)、绝对票息(trigger=True)的欧式雪球，可以作为DCN、凤凰的派息拆分的子结构(敲入价=派息线)，用PDE/QUAD定价"""

    def __init__(self, s0, barrier_out, barrier_in, coupon_out, trigger=False, strike_lower=None, margin_lvl=1,
                 coupon_div=None, lock_term=3, parti_in=1, obs_dates=None, pay_dates=None,
                 engine=None, status=StatusType.NoTouch, maturity=None, start_date=None, end_date=None,
                 trade_calendar=CN_CALENDAR, annual_days=AnnualDays.N365, t_step_per_year=243,
                 stoch_process=None, s=None, r=None, q=None, vol=None, engine_method=None):
        """继承自动赎回基类AutocallableBase的参数，详见AutocallableBase的__init__方法
        Args: 以下为区别于经典雪球的参数
            trigger: bool，是否为触发器，即敲出票息是否年化，True为绝对值，False为百分比
            strike_lower: float，敲入发生后(熊市价差)的低行权价，即保底边界，越高提供的保护越多，值为0时不保底
            margin_lvl: float，保证金比例，默认为1，即无杠杆。
                                不追保雪球一般将保证金比例设为最大可能亏损，例如75敲入，保证金比例25%，即为4倍杠杆
            parti_in: float，敲入后参与率，限损
        """
        super().__init__(in_obs_type=ExerciseType.European,  # 属性：到期观察敲入(欧式)
                         s0=s0, maturity=maturity, start_date=start_date, end_date=end_date, lock_term=lock_term,
                         trade_calendar=trade_calendar, obs_dates=obs_dates, pay_dates=pay_dates, status=status,
                         annual_days=annual_days, parti_in=parti_in, margin_lvl=margin_lvl,
                         t_step_per_year=t_step_per_year, engine=engine, stoch_process=stoch_process,
                         s=s, r=r, q=q, vol=vol, trigger=trigger, engine_method=engine_method)

        len_obs_dates = len(self.obs_dates.date_schedule) if isinstance(self.obs_dates.date_schedule,
                                                                        (list, np.ndarray)) else 1
        if isinstance(barrier_out, (int, float, np.int32, np.float64)):
            self.barrier_out = np.ones(len_obs_dates) * barrier_out
        elif isinstance(barrier_out, (list, np.ndarray)):
            self.barrier_out = barrier_out
        else:
            raise ValueError(f'敲出障碍价类型为{type(barrier_out)}，仅支持int/float/list/np.ndarray，请检查')
        if isinstance(barrier_in, (int, float, np.int32, np.float64)):
            self.barrier_in = barrier_in
        else:
            raise ValueError(f'敲入障碍价类型为{type(barrier_in)}，仅支持int/float，请检查')
        if isinstance(coupon_out, (int, float, np.int32, np.float64)):
            self.coupon_out = np.ones(len_obs_dates) * coupon_out
        elif isinstance(coupon_out, (list, np.ndarray)):
            self.coupon_out = coupon_out
        else:
            raise ValueError(f'敲出票息类型为{type(coupon_out)}，仅支持int/float/list/np.ndarray，请检查')
        self.coupon_div = coupon_div if coupon_div is not None else self.coupon_out[-1]  # 红利票息
        self.parti_out = 0  # 没有敲出上涨参与率
        self.strike_upper = s0  # 敲入后高行权价，默认s0
        self.strike_lower = strike_lower if strike_lower is not None else 0  # 敲入后低行权价（保底价）
        # 欧式雪球继承了父类的反算票息方法，但是只支持红利票息=敲出票息=常数的情形

    def __repr__(self):
        """返回期权的描述"""
        return "欧式雪球"



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import numpy as np
from pricelib.common.utilities.enums import StatusType
from pricelib.common.pricing_engine_base import QuadEngine
from pricelib.common.date_handler import global_evaluation_date
from .quad_snowball_engine import QuadSnowballEngine


class QuadEuroSnowEngine(QuadSnowballEngine):
    """欧式雪球 数值积分定价引擎
    由于仅到期观察敲入，只需要按敲出观察日反向迭代，重写父类的_data_preprocess和_backward_induction方法
    """

    def _data_preprocess(self, prod, calculate_date, spot, vol):
        """预处理：生成反向迭代的步数、fft的对数价格向量及边界的对数价格向量
        每日观察敲入雪球：
            网格在时间方向的步数是按逐个交易日迭代，还会生成每个交易日对应的下一个敲出观察日对应的付息日
        欧式雪球(仅到期观察敲入): 【子类QuadEuroSnowEngine对此方法重写】
            网格在时间方向的步数是按敲出观察日迭代
        """
        # 生成反向迭代的步数
        if 0 in self.out_dates:  # 如果估值日就是敲出观察日，则积分迭代步数少一次
            self.backward_steps = len(self.out_dates) - 1
        else:
            self.backward_steps = len(self.out_dates)

        # 初始化fft的对数价格向量及边界的对数价格向量
        self.init_grid(spot, vol, self._maturity_business_days / prod.t_step_per_year)
        self.s_vec = np.exp(self.ln_s_vec)
        self.j_vec = np.linspace(0, self.backward_steps, self.backward_steps + 1)  # 生成时间格点
        self.v_not_in = np.zeros(shape=(self.n_points, self.backward_steps + 1))
        self.v_knock_in = np.zeros(shape=(self.n_points, self.backward_steps + 1))

        # 最后一次迭代的时间步长: 从第一个敲出观察日到估值日的时间差, 用于_last_step_single_pv方法
        if self.out_dates[-1] == 0 and len(self.out_dates) > 1:  # 如果估值日就是敲出观察日
            self.last_time_step = self.out_dates[-2] / prod.t_step_per_year  # 最后一步的时间差是下一个敲出观察日
        else:
            self.last_time_step = self.out_dates[-1] / prod.t_step_per_year

        # 每个迭代日就是敲出观察日，对应的付息日与观察日的自然日年化之差
        self.next_diff_obspaydate = self.diff_obs_pay_dates  # 计息时间数

        # 每个迭代日就是敲出观察日，对应的票息支付日
        self.next_paydate = self.pay_dates  # 计息时间数

    def _backward_induction(self):
        """从倒数第一个观察日向第一个观察日逆向迭代"""
        prod = self.prod
        _out_dates = np.flip(self.out_dates) / prod.t_step_per_year
        if 0 in _out_dates:  # 如果估值日就是敲出观察日，则积分迭代步数少一次
            dt_vec = np.diff(_out_dates)[1:]
        else:
            dt_vec = np.diff(_out_dates)
        # 按照敲出观察日逆向迭代，迭代的第一次是倒数第二个观察日，迭代的最后一次是正数第一个观察日
        for j in range(self.backward_steps - 1, 0, -1):
            self.v_not_in[:, j] = self.fft_step_backward(self.ln_s_vec, self.ln_s_vec,
                                                         self.v_not_in[:, j + 1], dt_vec[j - 1])
            # 敲出票息是否为绝对收益
            coupon_t = 1 if prod.trigger else self.next_paydate[j - 1]
            # 如果敲出价是浮动的，调整敲出价
            self.next_barrier_out = (self._barrier_out[j - 1]
                                     if isinstance(prod.barrier_out, (list, np.ndarray)) else prod.barrier_out)
            self.out_idxs = np.array(np.where(self.s_vec >= self.next_barrier_out)[0])
            # 如果敲出票息是浮动的，调整敲出票息
            self.next_coupon_out = (self._coupon_out[j - 1]
                                    if isinstance(prod.coupon_out, (list, np.ndarray)) else prod.coupon_out)
            # 发生敲出的payoff
            knock_out_payoff = self.itm[self.out_idxs] + (
                prod.margin_lvl + self.next_coupon_out * coupon_t) * prod.s0 * self.process.interest.disc_factor(
                self.next_paydate[j - 1], self.next_paydate[j - 1] - self.next_diff_obspaydate[j - 1])
            # 用knock_out_payoff覆盖敲出价上方的衍生品价值
            self.v_not_in[self.out_idxs, j] = self.v_knock_in[self.out_idxs, j] = knock_out_payoff






#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import datetime
from pricelib.common.date_handler import CN_CALENDAR, AnnualDays, global_evaluation_date
from pricelib.common.utilities.utility import time_this
from pricelib.common.product_base import OptionBase
from pricelib.pricing_engines.analytic_engines.cashflow_engine import CashFlowEngine

__all__ = ['EnhancedIndexing']


class EnhancedIndexing(OptionBase):
    """指数增强"""

    def __init__(self, excess_return: float, s0: float, margin_lvl=1.,
                 start_date=None, end_date=None, maturity=None, *, engine=None,
                 trade_calendar=CN_CALENDAR, annual_days=AnnualDays.N365, t_step_per_year=243,
                 stoch_process=None, s=None, r=None, q=None, vol=None, engine_method=None):
        """构造函数
        Args:
            excess_return: float，年化的超额收益率
            s0: float，标的初始价格
            margin_lvl: float，保证金比例，默认为1，即无杠杆
            engine: 定价引擎，PricingEngine类，任意，因指数增强定价时，不需要定价引擎参与计算，只需要定价引擎中的标的资产随机过程
        时间参数: 要么输入年化期限，要么输入起始日和到期日；敲出观察日和票息支付日可缺省
            maturity: float，年化期限
            start_date: datetime.date，起始日
            end_date: datetime.date，到期日
            trade_calendar: 交易日历，Calendar类，默认为中国内地交易日历
            annual_days: int，每年的自然日数量
            t_step_per_year: int，每年的交易日数量
        可选参数:
            若未提供引擎的情况下，提供了标的价格、无风险利率、分红/融券率、波动率，
            则默认使用定价引擎 CashFlowEngine
            stoch_process: 随机过程StochProcessBase对象
            s: float，标的价格
            r: float，无风险利率
            q: float，分红/融券率
            vol: float，波动率
        """
        super().__init__()
        self.s0 = s0  # 标的初始价格
        self.trade_calendar = trade_calendar  # 交易日历
        self.annual_days = annual_days
        self.t_step_per_year = t_step_per_year
        assert maturity is not None or (start_date is not None and end_date is not None), "Error: 到期时间或起止时间必须输入一个"
        self.start_date = start_date if start_date is not None else global_evaluation_date()  # 起始时间
        self.end_date = end_date if end_date is not None else (
                self.start_date + datetime.timedelta(days=round(maturity * annual_days.value)))  # 结束时间
        if self.trade_calendar.isholiday(self.end_date):  # 如果到期日不是交易日，向后顺延一个交易日
            self.end_date = self.trade_calendar.advance(self.end_date, datetime.timedelta(days=1))
        if maturity is None:
            self.maturity = (end_date - start_date).days / annual_days.value
        else:
            self.maturity = maturity

        self.excess_return = excess_return
        self.margin_lvl = margin_lvl

        if engine is not None:
            self.set_pricing_engine(engine)
        elif stoch_process is not None:
            default_engine = CashFlowEngine(stoch_process)
            self.set_pricing_engine(default_engine)
        elif s is not None and r is not None and q is not None and vol is not None:
            default_engine = CashFlowEngine(s=s, r=r, q=q, vol=vol)
            self.set_pricing_engine(default_engine)

    def set_pricing_engine(self, engine):
        """设置定价引擎"""
        self.engine = engine

    def __repr__(self):
        """返回产品的描述"""
        return "指数增强"

    @time_this
    def price(self, t: datetime.date = None, spot=None):
        """计算指数增强估值
        Args:
            t: datetime.date，计算指数增强估值的日期
            spot: float/np.ndarray，标的价格
        Returns: 指数增强估值
        """
        calculate_date = global_evaluation_date() if t is None else t
        self.validate_parameters(t=t)
        # 当估值日就是到期日时
        if calculate_date == self.end_date:
            value = self.s0 * (self.margin_lvl + self.maturity * self.excess_return) + (spot - self.s0)
        else:
            tau = (self.end_date - calculate_date).days / self.annual_days  # 计算估值日到到期日的自然日期限
            r = self.engine.process.interest(tau)
            value = (self.s0 * (self.margin_lvl + self.maturity * self.excess_return)
                     + (spot * (1 + r * tau) - self.s0)) * self.engine.process.interest.disc_factor(tau)
        return value

    # def delta(self, t: datetime.date = None, *args, **kwargs):
    #     calculate_date = global_evaluation_date() if t is None else t
    #     tau = (self.end_date - calculate_date).days  # 计算估值日到到期日的自然日期限
    #     r = self.process.interest(tau)
    #     delta = (1 + r * tau)
    #     return delta

    def gamma(self, *args, **kwargs):
        gamma = 0
        return gamma

    def vega(self, *args, **kwargs):
        vega = 0
        return vega

    # def theta(self, t: datetime.date = None, *args, **kwargs):
    #     calculate_date = global_evaluation_date() if t is None else t
    #     theta = self.price() * -self.engine.process.interest(calculate_date)
    #     return theta
    #
    # def rho(self, *args, **kwargs):
    #     rho = 0
    #     return rho



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import numpy as np
from pricelib.common.date_handler import global_evaluation_date
from pricelib.common.pricing_engine_base import FdmGrid
from pricelib.common.utilities.enums import StatusType
from .fdm_autocallable_engine import FdmAutoCallableEngine


class FdmPhoenixEngine(FdmAutoCallableEngine):
    """ 凤凰类 AutoCallable PDE 有限差分法定价引擎：
            较雪球式AutoCallable，增加派息价格，少了敲出和红利票息
            每个派息（敲出）观察日，如果价格高于派息价格，派固定利息，如果发生敲出，合约提前结束；
            发生敲入后，派息方式不变，到期如果未敲出，结构为看跌空头或熊市价差空头
        FCN(Fixed Coupon Note, 固定票息票据)/DCN(Dynamic Coupon Note, 动态票息票据)/Phoenix凤凰票据
            FCN：无派息价格（派息价格为0），每月固定派息；到期日观察是否敲入
            DCN：有派息价格，每个观察日如果价格高于派息价格，派固定利息；到期日观察是否敲入
            Phoenix：有派息价格，每个观察日如果价格高于派息价格，派固定利息；每日观察是否敲入
    """

    def calc_present_value(self, prod, t=None, spot=None):
        """计算现值
        Args:
            prod: Product产品对象
            t: datetime.date，估值日; 如果是None，则使用全局估值日globalEvaluationDate
            spot: float，估值日标的价格，如果是None，则使用随机过程的当前价格
        Returns: float，现值
        """
        self.prod = prod
        calculate_date = global_evaluation_date() if t is None else t
        _maturity_business_days = prod.trade_calendar.business_days_between(calculate_date,
                                                                            prod.end_date) / prod.t_step_per_year
        if spot is None:
            spot = self.process.spot()

        self.t_step = prod.trade_calendar.business_days_between(calculate_date, prod.end_date)
        self.dt = 0 if self.t_step == 0 else 1 / prod.t_step_per_year  # 时间步长
        self.t_vec = np.linspace(0, _maturity_business_days, self.t_step + 1)

        if self.t_step == 0:  # 如果估值日是到期日
            self.smax = smin = spot  # 价格网格上下界都取估值日标的价格 - 直接计算到期日payoff
        else:
            smin = 0
            self.smax = self.n_smax * prod.s0  # 价格网格上界, 默认为n_smax倍初始价格
        # 初始化已敲入、未敲入的FdmGrid对象

        self.fd_not_in = FdmGrid(smax=self.smax, maturity=_maturity_business_days,
                                 t_step_per_year=prod.t_step_per_year, s_step=self.s_step,
                                 smin=smin)  # 未敲入
        self.fd_knockin = FdmGrid(smax=self.smax, maturity=_maturity_business_days,
                                  t_step_per_year=prod.t_step_per_year, s_step=self.s_step,
                                  smin=smin)  # 已敲入
        # 截取子产品（选取end_date大于calculate_date的产品）
        sub_products = [sub_product for sub_product in prod.sub_products if sub_product.end_date >= calculate_date]
        # 子产品定价
        for i, sub_product in enumerate(sub_products):
            sub_engine = sub_product.engine
            sub_engine.s_step = self.s_step
            sub_engine.n_smax = self.n_smax
            sub_engine.fdm_theta = self.fdm_theta
            sub_value = sub_product.price(t=t, spot=spot)
            # logging.info(f"{i}, {sub_product.__class__.__name__}, {sub_value}")
            """ np.pad
            将矩阵 pde_engine.fd_not_in.v_grid 的列数扩展到与 self.fd_not_in.v_grid 相同的宽度（通过右侧填充零），
            然后将这个扩展后的矩阵加到 mat1上。
                第一个元组 (0, 0)：表示在行方向（第一个维度）不进行填充
                第二个元组 (0, t_step + 1 - column_num)：表示在列方向（第二个维度）：
                    左侧填充 0 列, 右侧填充 t_step + 1 - column_num列
            使用常数填充模式, 填充值为 0
            """
            if sub_product.__class__.__name__ == "DigitalOption":
                # 二元期权子产品
                column_num = sub_engine.fdm.v_grid.shape[1]
                self.fd_not_in.v_grid += np.pad(sub_engine.fdm.v_grid, ((0, 0), (0, self.t_step + 1 - column_num)),
                                                mode='constant', constant_values=0)
                self.fd_knockin.v_grid += np.pad(sub_engine.fdm.v_grid, ((0, 0), (0, self.t_step + 1 - column_num)),
                                                 mode='constant', constant_values=0)
            else:
                # 欧式雪球或限亏雪球子产品
                column_num = sub_engine.fd_not_in.v_grid.shape[1]
                self.fd_not_in.v_grid += np.pad(sub_engine.fd_not_in.v_grid,
                                                ((0, 0), (0, self.t_step + 1 - column_num)), mode='constant',
                                                constant_values=0)
                if sub_product.__class__.__name__ == "EuropeanSnowball":
                    self.fd_knockin.v_grid += np.pad(sub_engine.fd_not_in.v_grid,
                                                     ((0, 0), (0, self.t_step + 1 - column_num)), mode='constant',
                                                     constant_values=0)
                elif sub_product.__class__.__name__ == "FlooredSnowball":
                    self.fd_knockin.v_grid += np.pad(sub_engine.fd_knockin.v_grid,
                                                     ((0, 0), (0, self.t_step + 1 - column_num)), mode='constant',
                                                     constant_values=0)
                else:
                    raise ValueError(
                        f"错误的子产品类型{sub_product.__class__.__name__}，应该是EuropeanSnowball或FlooredSnowball")

        if prod.status == StatusType.DownTouch:
            result = self.fd_knockin.functionize(self.fd_knockin.v_grid[1:-1, 0], kind="cubic")(spot)
        else:
            result = self.fd_not_in.functionize(self.fd_not_in.v_grid[1:-1, 0], kind="cubic")(spot)
        return np.float64(result)

    def _backward_induction(self):
        pass

    def _init_boundary_condition(self):
        pass

    def _init_terminal_condition(self):
        pass



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import numpy as np
from scipy.stats import norm
from pricelib.common.utilities.enums import CallPut, PaymentType, ExerciseType
from pricelib.common.date_handler import global_evaluation_date
from pricelib.common.pricing_engine_base import AnalyticEngine


class AnalyticAssetOrNothingEngine(AnalyticEngine):
    """二元(数字)期权-资产或无(asset or nothing)-闭式解定价引擎
    由于到期获得的payoff是标的资产多头，因此rebate固定为0"""

    # pylint: disable=invalid-name
    def calc_present_value(self, prod, t=None, spot=None):
        """计算现值
        Args:
            prod: Product产品对象
            t: datetime.date，估值日; 如果是None，则使用全局估值日globalEvaluationDate
            spot: float，估值日标的价格，如果是None，则使用随机过程的当前价格
        Returns: float，现值
        """
        assert prod.rebate == 0, "资产或无期权(asset or nothing)的rebate必须为0"
        calculate_date = global_evaluation_date() if t is None else t
        tau = prod.trade_calendar.business_days_between(calculate_date, prod.end_date) / prod.t_step_per_year
        if spot is None:
            spot = self.process.spot()

        if tau == 0:  # 如果估值日就是到期日，直接计算payoff
            if prod.callput == CallPut.Call:  # 向上
                value = spot if spot >= prod.strike else 0
                return value
            elif prod.callput == CallPut.Put:  # 向下
                value = spot if spot <= prod.strike else 0
                return value
            else:
                raise ValueError("Invalid callput type")

        r = self.process.interest(tau)
        q = self.process.div(tau)
        vol = self.process.vol(tau, spot)

        numerator = r - q + vol ** 2 / 2
        denominator = vol * np.sqrt(tau)
        mu = (r - q) / vol ** 2 - 0.5
        # 欧式
        if prod.exercise_type == ExerciseType.European:
            K_S = prod.strike / spot
            return spot * np.exp(-q * tau) * norm.cdf(
                prod.callput.value * (numerator * tau - np.log(K_S) / denominator))
        if prod.exercise_type == ExerciseType.American:
            if prod.discrete_obs_interval is not None:
                # 均匀离散观察障碍期权，M. Broadie, P. Glasserman, S.G. Kou(1997) 在连续障碍期权解析解上加调整项，调整障碍价格水平
                # 指数上的beta = -zeta(1/2) / sqrt(2*pi) = 0.5826, 其中zeta是黎曼zeta函数
                if prod.callput == CallPut.Call:  # 向上
                    strike = prod.strike * np.exp(0.5826 * vol * np.sqrt(prod.discrete_obs_interval))
                elif prod.callput == CallPut.Put:  # 向下
                    strike = prod.strike * np.exp(-0.5826 * vol * np.sqrt(prod.discrete_obs_interval))
                else:
                    raise ValueError("Invalid callput type")
            else:  # 连续观察
                strike = prod.strike
            K_S = strike / spot
            # 美式，立即支付
            if prod.payment_type == PaymentType.Hit:
                lambda_ = np.sqrt(mu ** 2 + 2 * r / vol ** 2)
                return strike * (K_S ** (mu + lambda_) * norm.cdf(
                    -prod.callput.value * (np.log(K_S) / denominator + lambda_ * denominator)) + K_S ** (
                                         mu - lambda_) * norm.cdf(
                    -prod.callput.value * (np.log(K_S) / denominator - lambda_ * denominator)))
            # 美式，到期支付
            elif prod.payment_type == PaymentType.Expire:
                return spot * np.exp(-q * tau) * (
                        norm.cdf((numerator * tau - np.log(K_S)) / denominator * prod.callput.value) +
                        K_S ** (2 * mu + 2) * norm.cdf(
                            -(numerator * tau + np.log(K_S)) / denominator * prod.callput.value))
            else:
                raise ValueError("Invalid payment type")
        # 观察方式既不是欧式也不是美式
        raise NotImplementedError(
            f"ExerciseType {prod.exercise_type.value}不支持的观察方式，仅支持欧式观察或美式观察.")



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import numpy as np
from pricelib.common.utilities.enums import CallPut, PaymentType, OutMechanism
from pricelib.common.pricing_engine_base import McEngine
from pricelib.common.date_handler import global_evaluation_date


class MCAccumulatorEngine(McEngine):
    """累购累沽 Monte Carlo 模拟定价引擎
    仅支持起初估值，不支持持仓估值，因为没有记录历史累积次数"""

    # pylint: disable=too-many-locals
    def calc_present_value(self, prod, t=None, spot=None):
        """计算现值
        Args:
            prod: Product产品对象
            t: datetime.date，估值日; 如果是None，则使用全局估值日globalEvaluationDate
            spot: float，估值日标的价格，如果是None，则使用随机过程的当前价格
        Returns: float，现值
                 如果输入spot = np.ndarray一维价格向量，t=终止日，则返回盈亏测算向量，即每个价格对应的现值，而不是平均值
        """
        calculate_date = global_evaluation_date() if t is None else t
        _maturity_business_days = prod.trade_calendar.business_days_between(calculate_date, prod.end_date)
        _obs_dates = np.array(prod.obs_dates.count_business_days(calculate_date))
        _pay_dates = np.array(prod.obs_dates.count_calendar_days(calculate_date))
        # 排除估值日=0（已实现的价格，已更新到status中）
        _obs_dates = _obs_dates[_obs_dates > 0]
        _pay_dates = _pay_dates[_pay_dates > 0]

        if spot is None:
            spot = self.process.spot()
        # 当估值日就是到期日时，直接用spot计算payoff，无需随机数矩阵
        if calculate_date == prod.end_date:
            if isinstance(spot, (list, np.ndarray)):  # spot是一个np.ndarray一维价格向量，返回盈亏测算向量，即每个价格对应的现值，而不是平均值
                paths = np.array([spot, ])
            else:  # spot是一个数值，计算期末payoff
                paths = np.array([[spot, ]])
        else:
            if spot != self.process.spot():
                self.reset_paths_flag()  # 重置路径标志位，重新生成路径
            paths = self.path_generator(n_step=_maturity_business_days, spot=spot,
                                        t_step_per_year=prod.t_step_per_year).copy()
        _n_path = paths.shape[1]

        # 敲出与行权价格
        strike = np.ones(_obs_dates.size) * prod.strike
        strike = np.tile(strike, (_n_path, 1)).T
        if prod.out_mechanism == OutMechanism.NoTerminate:  # 敲出不终止（标准累购累沽）
            barrier_out = np.ones(_obs_dates.size) * prod.barrier_out
            barrier_out = np.tile(barrier_out, (_n_path, 1)).T
            if prod.callput == CallPut.Call:  # 累购
                # 所有不利方向突破执行价的日期的布尔索引
                knock_in_time_idx = paths[_obs_dates, :] < strike
                # 1倍折价建仓的布尔索引
                no_touch_time_idx = np.where(paths[_obs_dates, :] > barrier_out, False, paths[_obs_dates, :] >= strike)
            else:  # 累沽
                # 所有不利方向突破执行价的日期的布尔索引
                knock_in_time_idx = paths[_obs_dates, :] > strike
                # 1倍折价建仓的布尔索引
                no_touch_time_idx = np.where(paths[_obs_dates, :] < barrier_out, False, paths[_obs_dates, :] <= strike)
            # 贴现因子(全部为到期日结算)
            discount_factor = self.process.interest.disc_factor(_pay_dates[-1] / prod.annual_days.value)
            if prod.payment_type == PaymentType.Expire:  # 若每个到期日建仓，发生敲出或到期时结算
                n_underlying = (np.sum(knock_in_time_idx, axis=0) * prod.leverage_ratio
                                + np.sum(no_touch_time_idx, axis=0)) * prod.amt + prod.status[0]
                # 到期盈亏 = 建仓量 * (到期S-K)
                payoff = ((prod.callput.value * (paths[-1] - prod.strike) * n_underlying) * discount_factor)
            elif prod.payment_type == PaymentType.Hit:  # 若每个观察日，立即以现金结算盈亏
                # 到期盈亏 = Σ{每个观察日的建仓量 * (该观察日S-K)}
                payoff = (((np.sum(knock_in_time_idx * (paths[_obs_dates, :] - prod.strike),
                                   axis=0) * prod.leverage_ratio
                            + np.sum(no_touch_time_idx * (paths[_obs_dates, :] - prod.strike),
                                     axis=0)) * prod.callput.value
                           * prod.amt) * discount_factor) + prod.status[0]
            else:
                raise ValueError(f"不支持的结算方式{prod.payment_type}")
        elif prod.out_mechanism == OutMechanism.Terminate:  # 敲出终止（熔断累购累沽）
            barrier_out = np.ones(_obs_dates.size) * prod.barrier_out
            barrier_out = np.tile(barrier_out, (_n_path, 1)).T
            if prod.callput == CallPut.Call:  # 累购
                # 最早发生敲出的日期的序号索引
                knock_out_time_idx = np.argmax(paths[_obs_dates, :] > barrier_out, axis=0)
                # 所有不利方向突破执行价的日期的布尔索引
                knock_in_time_idx = paths[_obs_dates, :] < strike
                # 1倍折价建仓的布尔索引
                no_touch_time_idx = np.where(paths[_obs_dates, :] > barrier_out, False, paths[_obs_dates, :] >= strike)
            else:  # 累沽
                # 最早发生敲出的日期的序号索引
                knock_out_time_idx = np.argmax(paths[_obs_dates, :] < barrier_out, axis=0)
                # 所有不利方向突破执行价的日期的布尔索引
                knock_in_time_idx = paths[_obs_dates, :] > strike
                # 1倍折价建仓的布尔索引
                no_touch_time_idx = np.where(paths[_obs_dates, :] < barrier_out, False, paths[_obs_dates, :] <= strike)
            # 每条路径，是否发生过敲出的布尔索引（0为未敲出）
            knock_out_scenario = np.where(knock_out_time_idx > 0, True, False)
            discount_time_idx = knock_out_time_idx.copy()
            discount_time_idx[~knock_out_scenario] = _obs_dates.size - 1  # 所有未敲出的路径，敲出时间设置为观察日期数量 - 1
            knock_out_time_idx[~knock_out_scenario] = _obs_dates.size  # 所有未敲出的路径，敲出时间设置为观察日期数量
            # 敲出终止（熔断），获取发生敲出之前的，不利方向突破执行价的日期的布尔索引
            knock_in_bool = np.where(np.arange(knock_in_time_idx.shape[0])[:, np.newaxis] > knock_out_time_idx,
                                     False, knock_in_time_idx)
            no_touch_bool = np.where(np.arange(no_touch_time_idx.shape[0])[:, np.newaxis] > knock_out_time_idx,
                                     False, no_touch_time_idx)
            # 贴现因子
            discount_factor = np.empty(_pay_dates.size)
            for i, j in enumerate(_pay_dates):
                discount_factor[i] = self.process.interest.disc_factor(j / prod.annual_days.value)
            # 敲出补偿：敲出每日固定收益 * 单日执行量 * 剩余敲出日数量（含敲出当日）
            knockout_rebate = prod.rebate * prod.amt * (_pay_dates.size - knock_out_time_idx)
            if prod.payment_type == PaymentType.Expire:  # 若每个到期日建仓，发生敲出或到期时结算
                n_underlying = (np.sum(knock_in_bool, axis=0) * prod.leverage_ratio
                                + np.sum(no_touch_bool, axis=0)) * prod.amt + prod.status[0]
                exercise_price = paths[_obs_dates[discount_time_idx], np.arange(_n_path)]
                # 到期盈亏 = 建仓量 * (到期S-K) + 敲出每日固定收益 * 单日执行量 * 剩余敲出日数量（含敲出当日）
                payoff = (prod.callput.value * (exercise_price - prod.strike) * n_underlying
                          + knockout_rebate) * discount_factor[discount_time_idx]
            elif prod.payment_type == PaymentType.Hit:  # 若每个观察日，立即以现金结算盈亏
                # 到期盈亏 = Σ{每个观察日的建仓量 * (该观察日S-K)} + 敲出每日固定收益 * 单日执行量 * 剩余敲出日数量（含敲出当日）
                payoff = (((np.sum(knock_in_bool * (paths[_obs_dates, :] - prod.strike), axis=0) * prod.leverage_ratio
                            + np.sum(no_touch_bool * (paths[_obs_dates, :] - prod.strike),
                                     axis=0)) * prod.callput.value * prod.amt
                           + knockout_rebate) * discount_factor[discount_time_idx]) + prod.status[0]
            else:
                raise ValueError(f"不支持的结算方式{prod.payment_type}")
        elif prod.out_mechanism == OutMechanism.NoKnockOut:  # 不敲出累购累沽

            if prod.callput == CallPut.Call:  # 累购
                # 所有不利方向突破执行价的日期的布尔索引
                knock_in_time_idx = paths[_obs_dates, :] < strike
                # 1倍折价建仓的布尔索引
                no_touch_time_idx = paths[_obs_dates, :] >= strike
            else:  # 累沽
                # 所有不利方向突破执行价的日期的布尔索引
                knock_in_time_idx = paths[_obs_dates, :] > strike
                # 1倍折价建仓的布尔索引
                no_touch_time_idx = paths[_obs_dates, :] <= strike
            # 贴现因子(全部为到期日结算)
            discount_factor = self.process.interest.disc_factor(_pay_dates[-1] / prod.annual_days.value)
            if prod.payment_type == PaymentType.Expire:  # 若每个到期日建仓，发生敲出或到期时结算
                n_underlying = (np.sum(knock_in_time_idx, axis=0) * prod.leverage_ratio
                                + np.sum(no_touch_time_idx, axis=0)) * prod.amt + prod.status[0]
                # 到期盈亏 = 建仓量 * (到期S-K)
                payoff = ((prod.callput.value * (paths[-1] - prod.strike) * n_underlying) * discount_factor)
            elif prod.payment_type == PaymentType.Hit:  # 若每个观察日，立即以现金结算盈亏
                # 到期盈亏 = Σ{每个观察日的建仓量 * (该观察日S-K)}
                payoff = (((np.sum(knock_in_time_idx * (paths[_obs_dates, :] - prod.strike),
                                   axis=0) * prod.leverage_ratio
                            + np.sum(no_touch_time_idx * (paths[_obs_dates, :] - prod.strike),
                                     axis=0)) * prod.callput.value
                           * prod.amt) * discount_factor) + prod.status[0]
            else:
                raise ValueError(f"不支持的结算方式{prod.payment_type}")
        else:
            raise ValueError("不支持的敲出机制: {}".format(prod.out_mechanism))

        if (calculate_date == prod.end_date) and isinstance(spot, np.ndarray):
            return payoff  # 终止日, spot是价格向量, 则返回盈亏测算向量，即每个价格对应的现值，而不是平均值
        else:
            return np.mean(payoff)



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import importlib_metadata
import importlib
from .pricing_engines import *
from .products import *
from .common import *

__author__ = '上海凌瓴信息科技有限公司: 马瑞祥, 夏鸿翔, 贡献: 张鹏任, 张峻尉'
__email__ = 'marx@galatech.com.cn'

try:  # 获取当前安装的包版本
    __version__ = importlib_metadata.version("pricelib")
except importlib_metadata.PackageNotFoundError:
    __version__ = "dev"


def check_for_updates(package_name):
    """检查PyPI是否有新版本的包可用"""
    if __version__ == "dev":
        logging.info(f"You are using a development version of {package_name}.")
        return
    try:  # 从PyPI获取最新版本信息
        requests = importlib.import_module("requests")
        response = requests.get(f"https://pypi.org/pypi/{package_name}/json")
        response.raise_for_status()
        data = response.json()
        latest_version = data['info']['version']
        if latest_version > __version__:
            logging.warning(f"A new version of {package_name} is available: {latest_version} "
                            f"(installed version: {__version__}),\nPlease update your package: "
                            f"pip install --upgrade {package_name} -i https://pypi.org/simple\n"
                            f"-------------------------------------------------------------------------------------")
        else:
            logging.info(f"You are using the latest version of {package_name} ({__version__}).")
    except Exception as e:
        logging.error(f"Failed to check for updates: {e},\nPlease check your internet connection and make sure "
                      f"you've already installed 'requests': pip install requests\n"
                      f"---------------------------------------------------------------------------------------")


# 当包被导入时自动检查更新
check_for_updates('pricelib')



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import math
import numpy as np
from pricelib.common.utilities.enums import CallPut, ExerciseType, PaymentType
from pricelib.common.pricing_engine_base import QuadEngine
from pricelib.common.date_handler import global_evaluation_date


class QuadDigitalEngine(QuadEngine):
    """二元(数字)期权-数值积分法定价引擎"""

    def calc_present_value(self, prod, t=None, spot=None, calc_greeks=False, s_step=0.001):
        """计算现值  todo 支持年化现金收益 rebate_annual
        Args:
            prod: Product产品对象
            t: datetime.date，估值日; 如果是None，则使用全局估值日globalEvaluationDate
            spot: float，估值日标的价格，如果是None，则使用随机过程的当前价格
            calc_greeks: bool，是否计算delta和gamma，默认为False，即只计算理论估值
            s_step: float，价格网格步长（百分比）
        Returns:
            若只计算pv，返回float，现值 pv
            若计算pv、delta和gamma，返回np.ndarray [down_price, pv, up_price] ，其中 down_price 和 up_price 是 s±Δs 的理论估值
        """
        calculate_date = global_evaluation_date() if t is None else t
        tau = prod.trade_calendar.business_days_between(calculate_date, prod.end_date) / prod.t_step_per_year
        if spot is None:
            spot = self.process.spot()

        if tau == 0:  # 估值日是到期日
            if calc_greeks:  # 计算理论估值、delta和gamma
                value = self._init_terminal_condition(prod, np.array([1 - s_step, 1, 1 + s_step]) * spot)
            else:  # 只计算理论估值, 直接计算终止条件
                value = self._init_terminal_condition(prod, np.array([spot]))[0]
            return value

        r = self.process.interest(tau)
        q = self.process.div(tau)
        vol = self.process.vol(tau, spot)
        dt = 1 / prod.t_step_per_year if prod.discrete_obs_interval is None else prod.discrete_obs_interval
        self.backward_steps = round(tau / dt)
        self._check_method_params()

        # 设置积分法engine参数
        self.set_quad_params(r=r, q=q, vol=vol)

        # 初始化fft的对数价格向量及边界的对数价格向量
        self.init_grid(spot, vol, tau)
        # 行权价在价格格点上的对应格点
        strike_point = np.where(np.exp(self.ln_s_vec) >= prod.strike)[0][0]
        v_vec = np.zeros(len(self.ln_s_vec))
        if prod.callput == CallPut.Call:
            v_vec[strike_point:] = prod.rebate
        elif prod.callput == CallPut.Put:
            v_vec[:strike_point] = prod.rebate
        # 欧式：
        if prod.exercise_type == ExerciseType.European:
            if calc_greeks:  # 计算理论估值、delta和gamma
                up_price = self.fft_step_backward(np.array([np.log(spot * (1 + s_step))]), self.ln_s_vec, v_vec, tau)[0]
                pv = self.fft_step_backward(np.array([np.log(spot)]), self.ln_s_vec, v_vec, tau)[0]
                down_price = self.fft_step_backward(np.array([np.log(spot * (1 - s_step))]), self.ln_s_vec, v_vec, tau)[0]
                value = np.array([down_price, pv, up_price])
            else:  # 只计算理论估值
                value = self.fft_step_backward(np.log(np.array([spot])), self.ln_s_vec, v_vec, tau)[0]
        # 美式：
        elif prod.exercise_type == ExerciseType.American:
            # 如果估值时的标的价格已经触碰行权价，直接返回行权收益的现值
            if (prod.callput == CallPut.Call and spot >= prod.strike) or (
                    prod.callput == CallPut.Put and spot <= prod.strike):
                rebate_pv = self.get_rebate_pv(prod, r, tau)
                if calc_greeks:  # 计算理论估值、delta和gamma
                    # 这里认为down_price, pv, up_price都是行权收益现值，忽略s±Δs与s跨越执行价的情形，即认为delta=gamma=0
                    return np.ones(3) * rebate_pv
                else:  # 只计算理论估值
                    return rebate_pv

            for step in range(self.backward_steps - 1, 0, -1):
                rebate_v = self.get_rebate_pv(prod, r, dt * (self.backward_steps - step))
                if prod.callput == CallPut.Call:
                    v_vec[:strike_point] = self.fft_step_backward(self.ln_s_vec[:strike_point], self.ln_s_vec,
                                                                  v_vec, dt)
                    v_vec[strike_point:] = rebate_v
                elif prod.callput == CallPut.Put:
                    v_vec[strike_point:] = self.fft_step_backward(self.ln_s_vec[strike_point:], self.ln_s_vec,
                                                                  v_vec, dt)
                    v_vec[:strike_point] = rebate_v
                else:
                    raise ValueError("Invalid call put type")
            if calc_greeks:  # 计算理论估值、delta和gamma
                up_price = self.fft_step_backward(np.array([np.log(spot * (1 + s_step))]), self.ln_s_vec, v_vec, dt)[0]
                pv = self.fft_step_backward(np.array([np.log(spot)]), self.ln_s_vec, v_vec, dt)[0]
                down_price = self.fft_step_backward(np.array([np.log(spot * (1 - s_step))]), self.ln_s_vec, v_vec, dt)[0]
                value = np.array([down_price, pv, up_price])
            else:  # 只计算理论估值
                value = self.fft_step_backward(np.array(np.log([spot])), self.ln_s_vec, v_vec, dt)[0]
        else:
            raise ValueError("ExerciseType must be American or European")
        return value

    @staticmethod
    def _init_terminal_condition(prod, s_vec):
        """获取终止时的期权价值，二元期权观察标的价格是否越过执行价格
        Args:
            prod: Product产品对象
            s_vec: np.ndarray, 网格的价格格点
        Returns:
            payoff: np.ndarray, 期末价值向量
        """
        if prod.callput == CallPut.Call:
            v_vec = np.where(s_vec >= prod.strike, prod.rebate, 0)
            return v_vec
        elif prod.callput == CallPut.Put:
            v_vec = np.where(s_vec <= prod.strike, prod.rebate, 0)
            return v_vec

    @staticmethod
    def get_rebate_pv(prod, r, tau):
        """计算行权收益的现值"""
        if prod.payment_type == PaymentType.Hit:
            rebate_v = prod.rebate
        elif prod.payment_type == PaymentType.Expire:
            rebate_v = prod.rebate * math.exp(-r * tau)
        else:
            raise ValueError("Invalid payment type")
        return rebate_v



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import numpy as np
import pandas as pd
from pandas.testing import assert_frame_equal
from examples.asian import asian_demo


def test_asian_demo():
    expected_data = [
        ['算术平均-代替标的结算价-Call亚式期权', 2.893737891784113, 2.9059603925765556, 2.9024890578353175],
        ['算术平均-代替标的结算价-Call增强型亚式期权', np.nan, 3.453030319689411, np.nan],
        ['算术平均-代替标的结算价-Put亚式期权', 4.34944253565525, 4.351327562759141, 4.358121279660046],
        ['算术平均-代替标的结算价-Put增强型亚式期权', np.nan, 4.898397489871997, np.nan],
        ['几何平均-代替标的结算价-Call亚式期权', 2.8091532462766367, 2.82014461530082, np.nan],
        ['几何平均-代替标的结算价-Call增强型亚式期权', np.nan, 3.3777986368091546, np.nan],
        ['几何平均-代替标的结算价-Put亚式期权', 4.474255200621862, 4.476119967804266, np.nan],
        ['几何平均-代替标的结算价-Put增强型亚式期权', np.nan, 5.008070041295551, np.nan],
        ['算术平均-代替执行价-Call亚式期权', np.nan, 2.8695188037729253, np.nan],
        ['算术平均-代替执行价-Put亚式期权', np.nan, 4.316900959109599, np.nan],
        ['几何平均-代替执行价-Call亚式期权', np.nan, 2.951289642124457, np.nan],
        ['几何平均-代替执行价-Put亚式期权', np.nan, 4.188063615140271, np.nan]]

    expected_df = pd.DataFrame(expected_data, columns=['类型', '闭式解', 'MonteCarlo', '二叉树'])
    res_df = asian_demo.run()
    assert_frame_equal(res_df, expected_df, check_dtype=True)



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import math
import numpy as np
from pricelib.common.date_handler import global_evaluation_date
from pricelib.common.utilities.enums import TouchType, ExerciseType, PaymentType
from pricelib.common.processes import StochProcessBase
from pricelib.common.pricing_engine_base.analytic_engine_base import (AnalyticEngine, euro_cash_or_nothing)
from pricelib.common.utilities.utility import logging


class AnalyticDoubleDigitalEngine(AnalyticEngine):
    """双边二元期权闭式级数近似解定价引擎
    欧式：二元凸式、二元凹式
        由欧式现金或无的解析解组合而成
    美式：美式双接触/双不接触
        Hui(1996) One-Touch Double Barrier Binary Option Values
        只支持美式双接触/双不接触, 到期支付；双接触期权的下边界payoff和上边界payoff必须相等；不支持年化的行权收益，只支持绝对数值的行权收益
        需要注意的是，Hui(1996)的形式是一个无穷级数，这个数列收敛非常迅速，大部分情况下前几项就足够确定期权价格，
                   但是在临近到期日时，这个解有误差，需要更多项级数来近似"""

    def __init__(self, stoch_process: StochProcessBase = None, series_num=10, *,
                 s=None, r=None, q=None, vol=None):
        """构造函数
        Args:
            stoch_process: 随机过程StochProcessBase对象
            series_num: int，计算Hui(1996)近似解时级数项的数量
        在未设置stoch_process时，(stoch_process=None)，会默认创建BSMprocess，需要输入以下变量进行初始化
            s: float，标的价格
            r: float，无风险利率
            q: float，分红/融券率
            vol: float，波动率
        """
        super().__init__(stoch_process=stoch_process, s=s, r=r, q=q, vol=vol)
        self.series_num = series_num
        self.i_vec = np.array(range(1, series_num + 1))  # 用于计算近似解的级数项

    # pylint: disable=invalid-name, too-many-locals
    def calc_present_value(self, prod, t=None, spot=None):
        """计算现值
        Args:
            prod: Product产品对象
            t: datetime.date，估值日; 如果是None，则使用全局估值日globalEvaluationDate
            spot: float，估值日标的价格，如果是None，则使用随机过程的当前价格
        Returns: float，现值
        """
        calculate_date = global_evaluation_date() if t is None else t
        _maturity_days = prod.trade_calendar.business_days_between(calculate_date, prod.end_date)
        _maturity = _maturity_days / prod.t_step_per_year
        tenor = (prod.stop_date - prod.start_date).days / prod.annual_days.value  # 从起始日到终止日的自然日时间（年）
        if spot is None:
            spot = self.process.spot()
        r = self.process.interest(_maturity)
        q = self.process.div(_maturity)
        vol = self.process.vol(_maturity, spot)

        if calculate_date == prod.stop_date:  # 如果估值日就是到期日，则直接计算payoff
            value = self.calc_maturity_payoff(prod, np.array([spot]), tenor)[0]
            return value

        if prod.exercise_type == ExerciseType.European:  # 如果是欧式双边二元期权（到期观察）
            return self._calc_european_double_digital(prod, spot, _maturity_days, _maturity, tenor, r, q, vol)
        elif prod.exercise_type == ExerciseType.American:  # 如果是美式双边二元期权
            assert prod.payment_type == PaymentType.Expire, "Error: 美式双边二元期权Hui(1996)只支持到期支付，请更换定价方法"
            return self._calc_american_double_digital(prod, spot, _maturity_days, _maturity, tenor, r, q, vol)
        else:
            raise ValueError("obs_type must be European or American")

    def calc_maturity_payoff(self, prod, s_vec, tenor):
        """到期时的期权价值，在障碍价格内侧，默认设定未敲入，未敲出
        Args:
            prod: Product产品对象
            s_vec: np.ndarray, 网格的价格格点
            tenor: float, 从起始日到终止日的自然日时间（年）
        Returns:
            payoff: np.ndarray, 期末价值向量
        """
        if prod.touch_type == TouchType.Touch:  # 二元凹式，此时相当于低行权价的现金或无put加上高行权价的现金或无call
            _rebate = self._preprocess_extra_returns(prod, tenor, prod.rebate)  # 高低触碰行权收益绝对数值
            v_vec = np.where(s_vec <= prod.bound[0], _rebate[0],
                             np.where(s_vec >= prod.bound[1], _rebate[1], 0))
        elif prod.touch_type == TouchType.NoTouch:  # 二元凸式，此时相当于低行权价的现金或无call减去高行权价的现金或无put
            _rebate = self._preprocess_extra_returns(prod, tenor, [prod.rebate[0], ])  # 中间的敲入现金返还绝对数值
            v_vec = np.where(s_vec > prod.bound[0], np.where(s_vec < prod.bound[1], _rebate[0], 0), 0)
        else:
            raise ValueError("touch_type must be Touch or NoTouch")
        return v_vec

    @staticmethod
    def _preprocess_extra_returns(prod, period, rebate_list):
        """预处理可能存在的年化现金返还rebate
        Args:
            prod: Product产品对象
            period: float 或 np.ndarray, 从起始到期权终止的时间（年）
                    对欧式障碍、美式敲入而言是整个存续期；对美式敲出而言，期权可能会因为敲出提前终止，此时period是起始日到敲出日的时间
            rebate_list: List[float], 现金返还率或绝对数值
        Returns:
            _rebate: float，处理后的现金返还绝对数值
        """
        if prod.rebate_annual:  # 年化百分比
            _rebate = []
            for r in rebate_list:
                _rebate.append(r * period * prod.s0)
        else:
            _rebate = rebate_list  # 绝对数值
        return _rebate

    def _calc_european_double_digital(self, prod, spot, _maturity_days, _maturity, tenor, r, q, vol):
        """二元凸式、二元凹式的解析解
            由欧式现金或无的解析解组合而成
        Args:
            prod: Product产品对象
            spot: float，估值日标的价格
            _maturity_days: float，交易日到期时间（天）
            _maturity: float，交易日到期时间（年）
            tenor: float, 从起始日到终止日的自然日时间（年）
            r: float，无风险利率（常数）
            q: float，连续分红率（常数）
            vol: float，波动率（常数）
        Returns:
            美式障碍期权的解析解
        """
        if prod.touch_type == TouchType.Touch:  # 二元凹式，此时相当于低行权价的现金或无put加上高行权价的现金或无call
            _rebate = self._preprocess_extra_returns(prod, tenor, prod.rebate)  # 高低触碰行权收益绝对数值
            value = (euro_cash_or_nothing(s=spot, k=prod.bound[0], t=_maturity, r=r, q=q, sigma=vol, sign=-1) *
                     _rebate[0] + euro_cash_or_nothing(s=spot, k=prod.bound[1], t=_maturity, r=r, q=q, sigma=vol,
                                                       sign=1) * _rebate[1])
        elif prod.touch_type == TouchType.NoTouch:  # 二元凸式，此时相当于低行权价的现金或无call减去高行权价的现金或无put
            _rebate = self._preprocess_extra_returns(prod, tenor, [prod.rebate[0], ])  # 中间的敲入现金返还绝对数值
            value = (euro_cash_or_nothing(s=spot, k=prod.bound[0], t=_maturity, r=r, q=q, sigma=vol, sign=1) -
                     euro_cash_or_nothing(s=spot, k=prod.bound[1], t=_maturity, r=r, q=q, sigma=vol, sign=1)
                     ) * _rebate[0]
        else:
            raise ValueError("touch_type must be Touch or NoTouch")
        return value

    def _calc_american_double_digital(self, prod, spot, _maturity_days, _maturity, tenor, r, q, vol):
        """美式双接触/双不接触的解析解
            Hui(1996) One-Touch Double Barrier Binary Option Values
            只支持到期支付的美式双接触/双不接触；并且，双接触期权的下边界payoff和上边界payoff必须相等
            需要注意的是，Hui(1996)的形式是一个无穷级数，这个数列收敛非常迅速，大部分情况下前几项就足够确定期权价格，
                       但是在临近到期日时，这个解有误差，需要更多项级数来近似
            支持Broadie, Glasserman和Kou(1995)提出的均匀离散观察障碍期权近似解.
        Args:
            prod: Product产品对象
            spot: float，估值日标的价格
            _maturity_days: float，交易日到期时间（天）
            _maturity: float，交易日到期时间（年）
            tenor: float, 从起始日到终止日的自然日时间（年）
            r: float，无风险利率（常数）
            q: float，连续分红率（常数）
            vol: float，波动率（常数）
        Returns:
            美式双接触/双不接触的解析解
        """
        _rebate = self._preprocess_extra_returns(prod, tenor, prod.rebate)  # 高低触碰行权收益/非触碰行权收益的绝对数值
        assert (prod.payment_type == PaymentType.Expire and
                prod.exercise_type == ExerciseType.American), "Error: Hui(1996)双边二元期权级数近似解, 只支持到期支付的美式双接触/双不接触"
        if prod.touch_type == TouchType.Touch:
            assert _rebate[0] == _rebate[1], "Error: Hui(1996)双边二元期权闭式解, 双接触期权的下边界payoff和上边界payoff必须相等"
        if _maturity_days <= 15 and self.series_num < 100000:
            # 在临近到期日时，较少的级数项会导致误差，需要更多项级数来近似
            logging.info(
                f"距离到期仅剩{_maturity_days}个交易日, 较少的级数项会导致误差，需要更多项级数来近似，已重设为100000项")
            self.series_num = 100000
            self.i_vec = np.array(range(1, self.series_num + 1))
        elif _maturity_days > 15:
            # 在非临近到期日时，数列收敛非常迅速，大部分情况下前几项就足够确定期权价格，设为10即可
            if self.series_num >= 100000:
                logging.info(
                    f"距离到期{_maturity_days}个交易日, 数列收敛非常迅速，大部分情况下前几项就足够确定期权价格，已重设为10项")
                self.series_num = 10
            self.i_vec = np.array(range(1, self.series_num + 1))

        # 如果估值时的标的价格已经触碰行权价，直接返回行权收益的现值
        if spot <= prod.bound[0] or spot >= prod.bound[1]:
            if prod.touch_type == TouchType.NoTouch:
                return 0
            elif prod.touch_type == TouchType.Touch:
                return _rebate[0] * math.exp(-r * _maturity)
            else:
                raise ValueError("Error: TouchType只能是NoTouch或Touch")

        if prod.discrete_obs_interval is not None:
            # 均匀离散观察，M. Broadie, P. Glasserman, S.G. Kou(1997) 在连续障碍期权解析解上加调整项，调整障碍价格水平
            # 指数上的beta = -zeta(1/2) / sqrt(2*pi) = 0.5826, 其中zeta是黎曼zeta函数
            bound = [None, None]
            bound[1] = prod.bound[1] * np.exp(0.5826 * vol * np.sqrt(prod.discrete_obs_interval))
            bound[0] = prod.bound[0] * np.exp(-0.5826 * vol * np.sqrt(prod.discrete_obs_interval))
        else:  # 连续观察
            bound = prod.bound

        Z = math.log(bound[1] / bound[0])
        alpha = - 0.5 * (2 * (r - q) / vol ** 2 - 1)
        beta = - 0.25 * (2 * (r - q) / vol ** 2 - 1) ** 2 - 2 * r / vol ** 2

        value0 = self.i_vec * np.pi / Z
        value1 = 2 * np.pi * self.i_vec * _rebate[0] / Z ** 2
        value2 = (spot / bound[0]) ** alpha - np.power(-1, self.i_vec) * (spot / bound[1]) ** alpha
        value3 = alpha ** 2 + value0 ** 2
        value4 = np.sin(value0 * np.log(spot / bound[0]))
        value5 = np.exp(-0.5 * vol ** 2 * _maturity * (value0 ** 2 - beta))
        value_series = value1 * value2 * value4 * value5 / value3  # 期权价值
        double_no_touch = np.sum(value_series)

        if prod.touch_type == TouchType.NoTouch:
            return double_no_touch
        if prod.touch_type == TouchType.Touch:
            return _rebate[0] * np.exp(-r * _maturity) - double_no_touch
        raise ValueError("Hui(1996)双边二元期权闭式解, 未知的触碰类型")



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import numpy as np
from pricelib.common.pricing_engine_base import McEngine
from pricelib.common.date_handler import global_evaluation_date


class MCPortfolioEngine(McEngine):
    """香草组合 Monte Carlo 模拟定价引擎"""

    def calc_present_value(self, prod, t=None, spot=None):
        """计算现值
        Args:
            prod: Product产品对象
            t: datetime.date，估值日; 如果是None，则使用全局估值日globalEvaluationDate
            spot: float，估值日标的价格，如果是None，则使用随机过程的当前价格
        Returns: float，现值
                        如果输入spot = np.ndarray一维价格向量，t=终止日，则返回盈亏测算向量，即每个价格对应的现值，而不是平均值
        """
        calculate_date = global_evaluation_date() if t is None else t
        _maturity = (prod.end_date - calculate_date).days / prod.annual_days.value
        n_step = round(_maturity * prod.t_step_per_year)
        if spot is None:
            spot = self.process.spot()
        # 当估值日就是到期日时，直接用spot计算payoff，无需随机数矩阵
        if calculate_date == prod.end_date:
            if isinstance(spot, (list, np.ndarray)):  # spot是一个np.ndarray一维价格向量，返回盈亏测算向量，即每个价格对应的现值，而不是平均值
                paths = np.array([spot, ])
            else:  # spot是一个数值，计算期末payoff
                paths = np.array([[spot, ]])
        else:
            if spot != self.process.spot():
                self.reset_paths_flag()  # 重置路径标志位，重新生成路径
            paths = self.path_generator(n_step=n_step, spot=spot, t_step_per_year=prod.t_step_per_year).copy()

        price = 0
        for vanilla, position in prod.vanilla_list:
            price += (np.maximum(vanilla.callput.value * (paths[-1, :] - vanilla.strike),
                                        0)) * position * self.process.interest.disc_factor(_maturity)

        if (calculate_date == prod.end_date) and isinstance(spot, np.ndarray):
            return price  # 终止日, spot是价格向量, 则返回盈亏测算向量，即每个价格对应的现值，而不是平均值
        else:
            return np.mean(price)
